from lifton import align, lifton_utils, lifton_class, get_id_fraction, variants, logger
import copy, os
from Bio.Seq import Seq
from intervaltree import Interval, IntervalTree

class Lifton_ORF:
    def __init__(self, start, end):
        self.start = start
        self.end = end


class Lifton_Status:
    def __init__(self):
        self.liftoff = 0
        self.miniprot = 0
        self.lifton_dna = 0
        self.lifton_aa = 0
        self.eval_dna = 0
        self.eval_aa = 0
        self.annotation = None
        self.status = []


class Lifton_Alignment:
    def __init__(self, extracted_identity, cds_children, alignment_query, alignment_comp, alignment_ref, cdss_protein_boundary, cdss_protein_aln_boundary, extracted_seq, reference_seq, db_entry):
        self.identity = extracted_identity
        self.cds_children = cds_children
        self.query_aln = alignment_query
        self.comp = alignment_comp
        self.ref_aln = alignment_ref
        self.cdss_protein_boundaries = cdss_protein_boundary
        self.cdss_protein_aln_boundaries = cdss_protein_aln_boundary
        self.query_seq = extracted_seq
        self.ref_seq = reference_seq
        self.db_entry = db_entry

    def write_alignment(self, outdir, tool_name, mutation, trans_id):
        outdir_tool = outdir+"/"+tool_name+"/"+mutation+"/"
        os.makedirs(outdir_tool, exist_ok=True)
        outfile= outdir_tool+trans_id+".fa"
        with open(outfile, "w") as fw:
            fw.write("> Reference\n")
            fw.write(self.ref_aln + "\n")
            fw.write("> Target\n")
            fw.write(self.query_aln + "\n")

    
class Lifton_feature:
    def __init__(self, id):
        self.id = id
        self.copy_num = 0
        self.is_protein_coding = False
        self.is_non_coding = False
        self.children = set()


class Lifton_GENE:
    def __init__(self, ref_gene_id, gffutil_entry_gene, ref_gene_attrs, tree_dict, ref_features_dict, args, tmp = False):
        ###########################
        # Assigning the reference gene & attributes
        ###########################
        self.entry = gffutil_entry_gene
        self.entry.source = "LiftOn"
        self.is_protein_coding = False
        self.is_non_coding = False
        self.transcripts = {}
        self.ref_gene_id = ref_gene_id
        self.copy_num = self.__get_gene_copy(ref_features_dict)
        self.tmp = tmp
        self.entry.attributes = ref_gene_attrs
        self.entry.attributes["ID"] = self.ref_gene_id + "_" + str(self.copy_num) if self.copy_num > 0 else self.ref_gene_id
        if self.copy_num > 0:
            self.entry.attributes["extra_copy_number"] = [str(self.copy_num)]        
        self.__update_gene_copy(ref_features_dict)
        self.entry.id = self.entry.attributes["ID"][0]
        gene_interval = Interval(self.entry.start, self.entry.end, self.entry.id)
        if self.entry.seqid not in tree_dict.keys():
            tree_dict[self.entry.seqid] = IntervalTree()
        tree_dict[self.entry.seqid].add(gene_interval)
        # Decide if its type
        # Check for gene type/biotype attribute in order of preference
        gene_type_key = None
        if args.annotation_database.upper() == "REFSEQ":
            # For RefSeq, check gene_biotype first, then biotype as fallback
            if "gene_biotype" in self.entry.attributes.keys():
                gene_type_key = "gene_biotype"
            elif "biotype" in self.entry.attributes.keys():
                gene_type_key = "biotype"
        elif args.annotation_database.upper() == "GENCODE" or args.annotation_database.upper() == "ENSEMBL" or args.annotation_database.upper() == "CHESS":
            # For GENCODE/ENSEMBL/CHESS, check gene_type first, then biotype as fallback
            if "gene_type" in self.entry.attributes.keys():
                gene_type_key = "gene_type"
            elif "biotype" in self.entry.attributes.keys():
                gene_type_key = "biotype"
        else:
            # For other databases, try all possible keys in order
            if "gene_biotype" in self.entry.attributes.keys():
                gene_type_key = "gene_biotype"
            elif "gene_type" in self.entry.attributes.keys():
                gene_type_key = "gene_type"
            elif "biotype" in self.entry.attributes.keys():
                gene_type_key = "biotype"
        
        if gene_type_key is not None:
            if self.entry.attributes[gene_type_key][0] == "protein_coding":
                self.is_protein_coding = True
            elif (self.entry.attributes[gene_type_key][0] == "lncRNA" or self.entry.attributes[gene_type_key][0] == "ncRNA"):
                self.is_non_coding = True
        
    def __get_gene_copy(self, ref_features_dict):
        if 'extra_copy_number' in self.entry.attributes:
            return int(self.entry.attributes['extra_copy_number'][0])
        else:
            if self.ref_gene_id in ref_features_dict.keys():
                return ref_features_dict[self.ref_gene_id].copy_num
            else:
                return 0

    def __update_gene_copy(self, ref_features_dict):
        if self.ref_gene_id in ref_features_dict.keys():
            ref_features_dict[self.ref_gene_id].copy_num += 1

    def update_gene_info(self, chromosome, start, end):
        self.entry.seqid = chromosome
        self.entry.start = start
        self.entry.end = end

    def add_miniprot_transcript(self, ref_trans_id, gffutil_entry_trans, ref_trans_attrs, ref_features_dict):
        Lifton_trans = Lifton_TRANS(ref_trans_id, self.ref_gene_id, self.entry.id, self.copy_num, gffutil_entry_trans, ref_trans_attrs)
        self.transcripts[Lifton_trans.entry.id] = Lifton_trans
        return Lifton_trans
    
    def update_trans_info(self, trans_id, chromosome, start, end):
        self.transcripts[trans_id].entry.seqid = chromosome
        self.transcripts[trans_id].entry.start = start
        self.transcripts[trans_id].entry.end = end

    def add_transcript(self, ref_trans_id, gffutil_entry_trans, ref_trans_attrs):
        Lifton_trans = Lifton_TRANS(ref_trans_id, self.ref_gene_id, self.entry.id, self.copy_num, gffutil_entry_trans, ref_trans_attrs)
        self.transcripts[Lifton_trans.entry.id] = Lifton_trans
        return Lifton_trans

    def add_feature(self, gffutil_entry_trans):
        Lifton_feature = LiftOn_FEATURE(self.entry.id, gffutil_entry_trans, self.copy_num)
        self.transcripts[Lifton_feature.entry.id] = Lifton_feature
        return Lifton_feature

    def add_exon(self, trans_id, gffutil_entry_exon):
        self.transcripts[trans_id].add_exon(gffutil_entry_exon)

    def add_cds(self, trans_id, gffutil_entry_cds):
        self.transcripts[trans_id].add_cds(gffutil_entry_cds)
                            
    def orf_search_protein(self, trans_id, ref_trans_id, fai, ref_proteins, fai_trans, lifton_status, eval_only=False, eval_liftoff_chm13=False):
        if trans_id not in self.transcripts.keys():
            return None, False
        if not eval_liftoff_chm13:
            ref_protein_seq = str(ref_proteins[ref_trans_id]) if ref_trans_id in ref_proteins.keys() else None
            ref_trans_seq = str(fai_trans[ref_trans_id]) if ref_trans_id in fai_trans.keys() else None
        else:
            ref_protein_seq = str(ref_proteins["rna-"+ref_trans_id]) if "rna-"+ref_trans_id in ref_proteins.keys() else None
            ref_trans_seq = str(fai_trans["rna-"+ref_trans_id]) if "rna-"+ref_trans_id in fai_trans.keys() else None
        lifton_aln, good_trans = self.transcripts[trans_id].orf_search_protein(fai, ref_protein_seq, ref_trans_seq, lifton_status, is_non_coding=self.is_non_coding, eval_only=eval_only)
        return lifton_aln, good_trans

    def update_cds_list(self, trans_id, cds_list):
        self.transcripts[trans_id].update_cds_list(cds_list)
        self.update_boundaries()

    def add_lifton_gene_status_attrs(self, source):
        self.entry.attributes["source"] = [source]

    def add_lifton_trans_status_attrs(self, trans_id, lifton_status):
        self.transcripts[trans_id].add_lifton_trans_status_attrs(lifton_status)

    def write_entry(self, fw, transcripts_stats_dict):
        if not self.tmp:
            fw.write(str(self.entry) + "\n")
        for key, trans in self.transcripts.items():
            trans.write_entry(fw)
            TYPE = ""
            if self.is_protein_coding and trans.entry.featuretype == "mRNA":
                TYPE = "coding"
            elif self.is_non_coding and (trans.entry.featuretype == "ncRNA" or trans.entry.featuretype == "nc_RNA" or trans.entry.featuretype == "lncRNA" or trans.entry.featuretype == "lnc_RNA"):
                TYPE = "non-coding"
            else:
                TYPE = "other"
            if trans.ref_tran_id is None:
                continue
            if not trans.ref_tran_id in transcripts_stats_dict[TYPE].keys():
                transcripts_stats_dict[TYPE][trans.ref_tran_id] = 1
            else:
                transcripts_stats_dict[TYPE][trans.ref_tran_id] += 1

    def update_boundaries(self):        
        for key, trans in self.transcripts.items():
            self.entry.start = trans.entry.start if trans.entry.start < self.entry.start else self.entry.start
            self.entry.end = trans.entry.end if trans.entry.end > self.entry.end else self.entry.end

    def print_gene(self):
        print(self.entry)
        for key, trans in self.transcripts.items():
            trans.print_transcript()


class LiftOn_FEATURE:
    def __init__(self, parent_id, gffutil_entry_feature, copy_num):
        self.entry = gffutil_entry_feature
        self.entry.source = "LiftOn"
        self.copy_num = copy_num
        self.features = {}
        self.entry.attributes["Parent"] = [parent_id]
        self.ref_tran_id = None
        # self.ref_tran_id = self.entry.id
        if int(copy_num) > 0:
            feature_id_base = lifton_utils.get_ID_base(self.entry.id)
            self.entry.id = f"{feature_id_base}_{copy_num}"
            self.entry.attributes["ID"] = [self.entry.id]

    def add_feature(self, gffutil_entry_trans):
        Lifton_feature = LiftOn_FEATURE(self.entry.id, gffutil_entry_trans, self.copy_num)
        self.features[Lifton_feature.entry.id] = Lifton_feature
        return Lifton_feature

    def write_entry(self, fw):
        fw.write(str(self.entry) + "\n")
        for key, feature in self.features.items():
            feature.write_entry(fw)

    def print_feature(self):
        print(self.entry)
        for key, feature in self.features.items():
            feature.print_feature()


class Lifton_TRANS:
    def __init__(self, ref_trans_id, ref_gene_id, gene_id, copy_num, gffutil_entry_trans, ref_trans_attrs):
        # Assigning the reference transcripts & attributes
        if int(copy_num) > 0:
            trans_id = f"{ref_trans_id}_{copy_num}" 
        else:
            trans_id = ref_trans_id
        self.entry = gffutil_entry_trans
        self.entry.attributes = ref_trans_attrs
        self.entry.id = trans_id
        self.entry.attributes["ID"] = [trans_id]
        self.entry.source = "LiftOn"
        self.exons = []
        self.exon_dic = {}
        # Get reference ID for the gene & transcript.
        self.ref_gene_id = ref_gene_id
        self.ref_tran_id = ref_trans_id
        if 'Parent' in self.entry.attributes:
            self.entry.attributes['Parent'] = [gene_id]
        if 'transcript_id' in self.entry.attributes:
            self.entry.attributes['transcript_id'] = [self.entry.id]

    def add_lifton_trans_status_attrs(self, lifton_status):
        self.entry.attributes["protein_identity"] = [f"{lifton_status.lifton_aa:.3f}"]
        self.entry.attributes["dna_identity"] = [f"{lifton_status.lifton_dna:.3f}"]
        self.entry.attributes["status"] = [lifton_status.annotation]

    def add_exon(self, gffutil_entry_exon):
        gffutil_entry_exon.attributes['Parent'] = [self.entry.id]
        Lifton_exon = Lifton_EXON(gffutil_entry_exon)
        lifton_utils.custom_bisect_insert(self.exons, Lifton_exon)

    def add_cds(self, gffutil_entry_cds):
        for exon in self.exons:
            _, ovp = lifton_utils.segments_overlap_length((exon.entry.start, exon.entry.end), (gffutil_entry_cds.start, gffutil_entry_cds.end))
            if ovp:
                gffutil_entry_cds.attributes['Parent'] = [self.entry.id]
                exon.add_cds(gffutil_entry_cds)

    def update_gffutil_entry_trans(self, gffutil_entry_trans):
        for key, atr in gffutil_entry_trans.attributes.items():
            self.entry.attributes[key] = atr

    def mv_exon_idx(self, exon_idx):
        if exon_idx < len(self.exons)-1:
            exon_idx += 1
        return exon_idx

    def update_cds_list(self, cds_list):
        idx_exon_itr = 0
        new_exons = []
        # Reverse CDS list if the strand is "-" => CDSs are in small to large order
        if self.entry.strand == "-":
            cds_list.reverse()
        ########################
        # Case 1: only 1 CDS 
        ########################
        if len(cds_list) == 1:
            only_cds = cds_list[0]
            processed_ovp_exons = False
            ovp_exons = []
            while idx_exon_itr < len(self.exons):
                exon = self.exons[idx_exon_itr]
                _, ovp = lifton_utils.segments_overlap_length((exon.entry.start, exon.entry.end), (only_cds.entry.start, only_cds.entry.end))

                # |eeeeee| |ccccc|
                if exon.entry.end < only_cds.entry.start:
                    new_exons.append(exon)
                elif ovp:
                    cp_exon = copy.deepcopy(exon)
                    ovp_exons.append(cp_exon)
                # |cccc|  |eeee|
                elif exon.entry.start > only_cds.entry.end: 
                    processed_ovp_exons = True
                    merged_exon = copy.deepcopy(exon)
                    if len(ovp_exons) == 0:
                        cp_exon = copy.deepcopy(exon)
                        ovp_exons.append(cp_exon)
                    merged_exon.entry.start = ovp_exons[0].entry.start
                    merged_exon.entry.end = ovp_exons[-1].entry.end
                    merged_exon.add_lifton_cds(only_cds)
                    new_exons.append(merged_exon)
                    new_exons.append(exon)
                idx_exon_itr += 1
            if not processed_ovp_exons:
                merged_exon = copy.deepcopy(exon)
                merged_exon.entry.start = ovp_exons[0].entry.start
                merged_exon.entry.end = ovp_exons[-1].entry.end
                merged_exon.add_lifton_cds(only_cds)
                new_exons.append(merged_exon)
        ########################
        # Case 2: only 1 exon, and >1 CDSs
        ########################
        elif len(self.exons) == 1:
            first_exon = self.exons[0]
            for cds_idx in range(len(cds_list)):
                exon = copy.deepcopy(first_exon)
                cds = cds_list[cds_idx]
                if cds_idx == 0:
                    if exon.entry.start >= cds.entry.start:
                        exon.entry.start = cds.entry.start
                    exon.entry.end = cds.entry.end
                elif cds_idx == len(cds_list)-1:
                    if exon.entry.end <= cds.entry.end:
                        exon.entry.end = cds.entry.end
                    exon.entry.start = cds.entry.start
                else:
                    exon.entry.end = cds.entry.end
                    exon.entry.start = cds.entry.start
                exon.add_lifton_cds(cds)
                new_exons.append(exon)
        ########################
        # Case 3: multiple CDSs and exons
        ########################
        elif len(cds_list) > 1 and len(self.exons) > 1:
            cds_idx = 0
            exon_idx = 0
            cds = cds_list[cds_idx]
            exon = self.exons[exon_idx]
            ################################################
            # Step 1: CDSs or exons are smaller & no overlapping 
            #      => finding the first overlapping CDS and exons
            #      => processing all exons / CDS ahead of the first overlapping
            ################################################
            init_head_order = False
            # |ccccc||eeeeee|
            if exon.entry.start >= cds.entry.end:
                init_head_order = False
            # |eeeeee| |ccccc|
            elif exon.entry.start < cds.entry.start:
                init_head_order = True
            if init_head_order:
                # Find the first overlapping exon with CDS
                while exon_idx < len(self.exons) and cds_idx < len(cds_list):
                    exon_start, exon_end = self.exons[exon_idx].entry.start, self.exons[exon_idx].entry.end
                    cds_start, cds_end = cds_list[cds_idx].entry.start, cds_list[cds_idx].entry.end
                    _, ovp = lifton_utils.segments_overlap_length((exon_start, exon_end), (cds_start, cds_end))
                    # |eeeeee| |ccccc|
                    if exon_end < cds_start:
                        # Create a new exon using exon boundary
                        new_exon = copy.deepcopy(self.exons[exon_idx])
                        new_exon.update_exon_info(exon_start, exon_end)
                        new_exons.append(new_exon)
                        exon_idx += 1
                    # |eeee|--|cccc|
                    elif ovp:
                        # Stop if exon and CDS overlap
                        # Create a new exon using exon boundary
                        new_exon = copy.deepcopy(self.exons[exon_idx])
                        new_exon.update_exon_info(min(exon_start, cds_start), cds_end)
                        new_exon.add_lifton_cds(cds_list[0])
                        new_exons.append(new_exon)
                        exon_idx += 1
                        cds_idx += 1
                        break
                    # |cccc|  |eeee|
                    elif exon_start > cds_end: 
                        # Create a new exon using CDS boundary
                        new_exon = copy.deepcopy(self.exons[exon_idx])
                        new_exon.update_exon_info(cds_start, cds_end)
                        new_exon.add_lifton_cds(cds_list[0])
                        new_exons.append(new_exon)
                        cds_idx += 1
                        break
            ################################################
            # Step 2: parse the inner exons and CDSs
            ################################################
            # Handle the CDSs in the middle
            # Leave the last CDS not processed.
            while cds_idx < len(cds_list)-1:
                cds = cds_list[cds_idx]
                # Create a new exon
                # Add the CDS in the new exon
                new_exon = copy.deepcopy(exon)
                new_exon.update_exon_info(cds.entry.start, cds.entry.end)
                new_exon.add_lifton_cds(cds)
                new_exons.append(new_exon)
                cds_idx += 1
            ################################################
            # Step 3: parse the last CDS & its overlapping exon
            ################################################
            last_cds_processed = False
            cds = cds_list[cds_idx]
            cds_start, cds_end = cds_list[cds_idx].entry.start, cds_list[cds_idx].entry.end
            while exon_idx < len(self.exons):
                exon = self.exons[exon_idx]
                exon_start, exon_end = exon.entry.start, exon.entry.end
                _, ovp = lifton_utils.segments_overlap_length((exon_start, exon_end), (cds_start, cds_end))
                # |eeeeee| |ccccc|
                if exon_end < cds_start:
                    # exon_idx = self.mv_exon_idx(exon_idx)
                    exon_idx += 1
                    continue
                elif ovp:
                    # Create a new exon using exon boundary
                    last_cds_processed = True
                    new_exon = copy.deepcopy(exon)
                    new_exon.update_exon_info(cds_start, max(cds_end, exon_end))
                    new_exon.add_lifton_cds(cds)
                    new_exons.append(new_exon)
                    exon_idx += 1
                    break
                # |cccc|  |eeee|
                elif exon_start > cds_end: 
                    # Create a new exon using CDS boundary
                    last_cds_processed = True
                    new_exon = copy.deepcopy(exon)
                    new_exon.update_exon_info(cds_start, cds_end)
                    new_exon.add_lifton_cds(cds)
                    new_exons.append(new_exon)
                    break
            ################################################
            # Step 4: parse the remaining exons
            ################################################
            while exon_idx < len(self.exons):
                exon = self.exons[exon_idx]
                # Create a new exon
                # Add the CDS in the new exon
                new_exon = copy.deepcopy(exon)
                new_exon.update_exon_info(new_exon.entry.start, new_exon.entry.end)
                new_exons.append(new_exon)
                exon_idx += 1
            ################################################
            # Step 5: parse the last CDS if it does not overlap with any exon
            ################################################
            if not last_cds_processed:
                # Create a new exon
                # Add the CDS in the new exon
                new_exon = copy.deepcopy(exon)
                new_exon.update_exon_info(cds.entry.start, cds.entry.end)
                new_exon.add_lifton_cds(cds)
                new_exons.append(new_exon)
        self.exons = new_exons
        self.update_boundaries()

    def get_coding_seq(self, fai):
        coding_seq = ""
        cdss_lens = []
        cds_children = []
        for exon in self.exons:
            if exon.cds is not None:
                cds_children.append(copy.deepcopy(exon.cds.entry))
                # Chaining the CDS features
                p_seq = exon.cds.entry.sequence(fai)
                if exon.cds.entry.strand == '-':
                    coding_seq = p_seq + coding_seq
                    cdss_lens.insert(0, exon.cds.entry.end - exon.cds.entry.start + 1)
                elif exon.cds.entry.strand == '+':
                    coding_seq = coding_seq + p_seq
                    cdss_lens.append(exon.cds.entry.end - exon.cds.entry.start + 1)
        return coding_seq, cds_children, cdss_lens

    def get_coding_trans_seq(self, fai):
        trans_seq = ""
        coding_seq = ""
        accum_cds_length = 0
        lcl_exons = []
        lcl_exons = self.exons
        if len(self.exons) > 0 and self.exons[0].entry.strand == '-':
            lcl_exons = self.exons[::-1]
        for exon in lcl_exons:
            # Chaining the exon features
            p_trans_seq = exon.entry.sequence(fai)
            p_trans_seq = Seq(p_trans_seq).upper()
            trans_seq = trans_seq + p_trans_seq
            if exon.cds is not None:
                # Updating CDS frame
                exon.cds.entry.frame = str(self.__get_cds_frame(accum_cds_length))
                accum_cds_length += (exon.cds.entry.end - exon.cds.entry.start + 1)
                # Chaining the CDS features
                p_seq = exon.cds.entry.sequence(fai)
                coding_seq = coding_seq + p_seq
        if trans_seq != None:
            trans_seq = str(trans_seq).upper()
        if coding_seq != None:
            coding_seq = str(coding_seq).upper()
        return coding_seq, trans_seq

    def translate_coding_seq(self, coding_seq):
        protein_seq = None
        if coding_seq != "":
            protein_seq = str(Seq(coding_seq).translate())
        return protein_seq

    def align_coding_seq(self, protein_seq, ref_protein_seq, lifton_status):
        if ref_protein_seq == "" or ref_protein_seq == None:
            lifton_aa_aln = None
            peps = None
        elif protein_seq == "" or protein_seq == None:
            lifton_aa_aln = None
            peps = None
        else:
            peps = protein_seq.split("*")
            lifton_aa_aln = align.protein_align(protein_seq, ref_protein_seq)
            # Update lifton sequence identity
            lifton_status.lifton_aa = max(lifton_status.lifton_aa, lifton_aa_aln.identity)
        return lifton_aa_aln, peps

    def align_trans_seq(self, trans_seq, ref_trans_seq, lifton_status):
        if ref_trans_seq == "" or ref_trans_seq == None:
            lifton_tran_aln = None
        elif trans_seq == "" or trans_seq == None:
            lifton_tran_aln = None
        else:
            lifton_tran_aln = align.trans_align(trans_seq, ref_trans_seq)
            lifton_status.lifton_dna = lifton_tran_aln.identity
        return lifton_tran_aln

    def orf_search_protein(self, fai, ref_protein_seq, ref_trans_seq, lifton_status, is_non_coding, eval_only=False):
        coding_seq, trans_seq = self.get_coding_trans_seq(fai)
        protein_seq = self.translate_coding_seq(coding_seq)
        # Aligning the LiftOn protein & DNA sequences
        lifton_aa_aln, peps = self.align_coding_seq(protein_seq, ref_protein_seq, lifton_status)
        lifton_tran_aln = self.align_trans_seq(trans_seq, ref_trans_seq, lifton_status)
        variants.find_variants(lifton_tran_aln, lifton_aa_aln, lifton_status, peps, is_non_coding)
        ORF_search = False
        for mutation in lifton_status.status:
            # identical # synonymous 
            # (1) inframe_insertion # (2) inframe_deletion # (3) nonsynonymous 
            # (4) frameshift # (5) start_lost # (6) stop_missing # (7) stop_codon_gain
            # Adding mutations in to entry.attributes
            if mutation != "identical":
                if "mutation" not in self.entry.attributes:
                    self.entry.attributes["mutation"] = [mutation]
                else:
                    self.entry.attributes["mutation"].append(mutation)
            # ORF searching for these four types of mutations
            # frameshift_orf_threshold = 0.8
            # and lifton_aa_aln.identity < frameshift_orf_threshold)
            if mutation == "stop_missing" or mutation == "stop_codon_gain" or mutation == "frameshift"  or mutation == "start_lost":
                ORF_search = True
        if ORF_search and eval_only==False:
            self.__find_orfs(trans_seq, ref_protein_seq, lifton_aa_aln, lifton_status)
        return lifton_tran_aln, lifton_aa_aln

    def __find_orfs(self, trans_seq, ref_protein_seq, lifton_aln, lifton_status):
        trans_seq = trans_seq.upper()
        # Find ORFs in whole transcript region (including UTRs)
        start_codon = "ATG"
        stop_codons = ["TAA", "TAG", "TGA"]
        orf_list = []
        max_orf_len = [0, 0, 0]
        for frame in range(3):
            orf_idx_s = 0
            for i in range(frame, len(trans_seq), 3):
                codon = str(trans_seq[i:i+3])
                if codon == start_codon:
                    orf_idx_s = i
                    orf_idx_e = i
                    orf_seq = ""
                    for j in range(i, len(trans_seq), 3):
                        codon = str(trans_seq[j:j+3])
                        orf_seq += codon
                        if codon in stop_codons:
                            orf_idx_e = j+3
                            break
                    if orf_seq and (orf_idx_s < orf_idx_e):
                        curr_orf_len = orf_idx_e-orf_idx_s+1
                        if curr_orf_len >  max_orf_len[frame]:
                            max_orf_len[frame] = curr_orf_len
                            orf = lifton_class.Lifton_ORF(orf_idx_s, orf_idx_e)
                            orf_list.append(orf)
        final_orf = None
        update_orf = False
        max_identity = 0
        for i, orf in enumerate(orf_list):
            orf_DNA_seq = trans_seq[orf.start:orf.end]
            orf_protein_seq = str(Seq(orf_DNA_seq).translate())
            orf_parasail_res = align.parasail_align_protein_base(orf_protein_seq, ref_protein_seq)
            orf_matches, orf_length = get_id_fraction.get_AA_id_fraction(orf_parasail_res.traceback.ref, orf_parasail_res.traceback.query)
            orf_identity = orf_matches/orf_length
            if orf_identity > max_identity:
                max_identity = orf_identity
                final_orf = orf
        # Only update orf if at least one frame similarity is larger than the original lifton_aa by the threshold
        threshold_orf = 0.01
        if max_identity > (lifton_status.lifton_aa + threshold_orf):
            lifton_status.lifton_aa = max_identity
            update_orf = True
        if final_orf is not None and update_orf:
            self.__update_cds_boundary(final_orf)

    def __update_cds_boundary(self, final_orf):
        if self.entry.strand == "+":
            self.__iterate_exons_update_cds(final_orf, self.exons, "+")
        elif self.entry.strand == "-":
            self.__iterate_exons_update_cds(final_orf, self.exons[::-1], "-")            

    def __iterate_exons_update_cds(self, final_orf, exons, strand):
        accum_exon_length = 0
        accum_cds_length = 0
        for exon_idx, exon in enumerate(exons):
            curr_exon_len = exon.entry.end - exon.entry.start + 1
            if accum_exon_length <= final_orf.start:
                if final_orf.start < accum_exon_length+curr_exon_len:
                    # Create first partial CDS
                    if exon.cds is not None:
                        if strand == "+":
                            exon.cds.entry.end = exon.entry.end
                            exon.cds.entry.start = exon.entry.start + (final_orf.start - accum_exon_length)
                        elif strand == "-":
                            exon.cds.entry.start = exon.entry.start 
                            exon.cds.entry.end = exon.entry.end - (final_orf.start - accum_exon_length)
                    else:
                        if strand == "+":
                            exon.add_novel_lifton_cds(exon.entry, exon.entry.start + (final_orf.start - accum_exon_length), exon.entry.end)
                        elif strand == "-":
                            exon.add_novel_lifton_cds(exon.entry, exon.entry.start, exon.entry.end - (final_orf.start - accum_exon_length))
                    exon.cds.entry.frame = str(self.__get_cds_frame(accum_cds_length))
                    accum_cds_length += (exon.cds.entry.end - exon.cds.entry.start + 1)
                else:
                    # No CDS should be created
                    exon.cds = None
            elif final_orf.start < accum_exon_length and accum_exon_length < final_orf.end:
                if final_orf.end <= accum_exon_length+curr_exon_len:
                    # Create the last partial CDS
                    if exon.cds is not None:
                        if strand == "+":
                            exon.cds.entry.start = exon.entry.start
                            exon.cds.entry.end = exon.entry.start + (final_orf.end - accum_exon_length)-1
                        elif strand == "-":
                            exon.cds.entry.end = exon.entry.end
                            exon.cds.entry.start = exon.entry.end - (final_orf.end - accum_exon_length)+1
                    else:
                        if strand == "+":
                            exon.add_novel_lifton_cds(exon.entry, exon.entry.start, exon.entry.start + (final_orf.end - accum_exon_length)-1)
                        elif strand == "-":
                            exon.add_novel_lifton_cds(exon.entry, exon.entry.end - (final_orf.end - accum_exon_length)+1, exon.entry.end)
                else:
                    # Keep the original full CDS / extend the CDS to full exon length
                    if exon.cds is None:
                        exon.add_novel_lifton_cds(exon.entry, exon.entry.start, exon.entry.end)
                    else:
                        exon.cds.update_CDS_info(exon.entry.start, exon.entry.end)
                exon.cds.entry.frame = str(self.__get_cds_frame(accum_cds_length))
                accum_cds_length += (exon.cds.entry.end - exon.cds.entry.start + 1)
            elif final_orf.end <= accum_exon_length:
                # No CDS should be created
                exon.cds = None
            accum_exon_length += curr_exon_len

    def __get_cds_frame(self, accum_cds_length):
        return (3 - accum_cds_length%3)%3

    def write_entry(self, fw):
        fw.write(str(self.entry) + "\n")
        # Write out the exons first
        for exon in self.exons:
            exon.write_entry(fw)
        # Write out the CDSs second
        for exon in self.exons:
            if exon.cds is not None:
                exon.cds.write_entry(fw)

    def update_boundaries(self):
        self.entry.start = self.exons[0].entry.start
        self.entry.end = self.exons[-1].entry.end

    def print_transcript(self):
        print(f"\t{self.entry}")
        for exon in self.exons:
            exon.print_exon()


class Lifton_EXON:
    def __init__(self, gffutil_entry_exon):
        gffutil_entry_exon.source = "LiftOn"
        gffutil_entry_exon.featuretype = "exon"
        self.entry = gffutil_entry_exon
        if 'extra_copy_number' in self.entry.attributes: self.entry.attributes.pop('extra_copy_number')
        self.cds = None

    def update_exon_info(self, start, end):
        self.cds = None
        self.entry.source = "LiftOn"
        self.entry.start = start
        self.entry.end = end

    def add_cds(self, gffutil_entry_cds):
        Lifton_cds = Lifton_CDS(gffutil_entry_cds)
        self.cds = Lifton_cds

    def add_novel_lifton_cds(self, gffutil_entry_exon, start, end):
        gffutil_entry_cds = copy.deepcopy(gffutil_entry_exon)
        gffutil_entry_cds.featuretype = "CDS"
        gffutil_entry_cds.start = start
        gffutil_entry_cds.end = end
        Lifton_cds = Lifton_CDS(gffutil_entry_cds)
        attributes = {}
        attributes['Parent'] = self.entry.attributes['Parent']
        Lifton_cds.entry.attributes = attributes
        self.cds = Lifton_cds

    def add_lifton_cds(self, Lifton_cds):
        if Lifton_cds is not None:
            attributes = {}
            attributes['Parent'] = self.entry.attributes['Parent']
            Lifton_cds.entry.attributes = attributes
        self.cds = Lifton_cds

    def write_entry(self, fw):
        fw.write(str(self.entry) + "\n")

    def print_exon(self):
        print(f"\t\t{self.entry}")
        if self.cds != None:
            self.cds.print_cds()


class Lifton_CDS:
    def __init__(self, gffutil_entry_cds):
        gffutil_entry_cds.source = "LiftOn"
        gffutil_entry_cds.featuretype = "CDS"
        self.entry = gffutil_entry_cds
        if 'extra_copy_number' in self.entry.attributes: self.entry.attributes.pop('extra_copy_number')

    def update_CDS_info(self, start, end):
        self.entry.source = "LiftOn"
        self.entry.start = start
        self.entry.end = end

    def write_entry(self, fw):
        fw.write(str(self.entry) + "\n")

    def print_cds(self):
        print(f"\t\t{self.entry}")


import copy 
from lifton import lifton_class, lifton_utils

def initialize_lifton_gene_eval(locus, ref_db, tree_dict, ref_features_dict, args, with_exons=False):
    """
        This function initializes Lifton gene instance.

        Parameters:
        - locus: gffutils feature instance
        - ref_db: reference database
        - tree_dict: intervaltree dictionary for each chromosome
        - ref_features_dict: reference features dictionary
        - with_exons: True if the gene has exons, False otherwise

        Returns:
        lifton_gene: Lifton gene instance
        ref_gene_id: reference gene
        ref_trans_id: reference transcript
    """
    ref_gene_id, ref_trans_id = lifton_utils.get_ref_ids_liftoff(ref_features_dict, locus.id, None)
    if ref_gene_id is None: return None, None, None
    if not args.evaluation_liftoff_chm13:
        lifton_gene = lifton_class.Lifton_GENE(ref_gene_id, copy.deepcopy(locus), copy.deepcopy(ref_db[ref_gene_id].attributes), tree_dict, ref_features_dict, args, tmp=with_exons)
    else:
        lifton_gene = lifton_class.Lifton_GENE(ref_gene_id, copy.deepcopy(locus), copy.deepcopy(ref_db["gene-"+ref_gene_id].attributes), tree_dict, ref_features_dict, args, tmp=with_exons)
    return lifton_gene, ref_gene_id, ref_trans_id


def lifton_add_trans_exon_cds_eval(lifton_gene, locus, ref_db, tgt_db, ref_trans_id, args):
    """
        This function adds transcript, exons, and CDSs to the Lifton gene instance.

        Parameters:
        - lifton_gene: Lifton gene instance
        - locus: gffutils feature instance
        - ref_db: reference database
        - tgt_db: liftoff feature database
        - ref_trans_id: reference transcript ID

        Returns:
        lifton_trans: Lifton transcript instance
        len(cdss_list): number of CDSs
    """
    try:
        ref_db["rna-"+ref_trans_id]
    except:
        return None, 0
    if not args.evaluation_liftoff_chm13:
        lifton_trans = lifton_gene.add_transcript(ref_trans_id, copy.deepcopy(locus), copy.deepcopy(ref_db[ref_trans_id].attributes))
    else:
        try:
            ref_db["rna-"+ref_trans_id]
        except:
            return None, 0
        lifton_trans = lifton_gene.add_transcript(ref_trans_id, copy.deepcopy(locus), copy.deepcopy(ref_db["rna-"+ref_trans_id].attributes))
    exons = tgt_db.children(locus, featuretype='exon', order_by='start')
    for exon in list(exons):
        lifton_gene.add_exon(lifton_trans.entry.id, exon)
    cdss = tgt_db.children(locus, featuretype=('CDS', 'stop_codon'), order_by='start') 
    cdss_list = list(cdss)
    for cds in cdss_list:
        lifton_gene.add_cds(lifton_trans.entry.id, cds)
    return lifton_trans, len(cdss_list)


def evaluation(lifton_gene, locus, ref_db, tgt_db, tree_dict, tgt_fai, ref_proteins, ref_trans, ref_features_dict, fw_score, args, ENTRY_FEATURE=False):
    """
        This function evaluates annotation.

        Parameters:
        - lifton_gene: Lifton gene instance
        - locus: feature instance 
        - ref_db: reference database
        - tgt_db: liftoff feature database
        - ref_id_2_m_id_trans_dict: reference id to miniprot transcript ids dictionary
        - m_feature_db: miniprot feature database
        - tree_dict: intervaltree dictionary for each chromosome
        - tgt_fai: target fasta index
        - ref_proteins: reference protein dictionary
        - ref_trans: reference transcript dictionary
        - ref_features_dict: reference features dictionary
        - fw_score: file writer for scores
        - fw_chain: file writer for chains
        - write_chains: write chains or not
        - DEBUG: debug mode
        - ENTRY_FEATURE: True if the feature is the root feature for a gene locus

        Returns:
        lifton_gene: LiftOn gene instance
    """
    exon_children = list(tgt_db.children(locus, featuretype='exon', level=1, order_by='start'))
    if lifton_gene is None and ENTRY_FEATURE:   
        # Gene (1st) features
        lifton_gene, ref_gene_id, ref_trans_id = initialize_lifton_gene_eval(locus, ref_db, tree_dict, ref_features_dict, args, with_exons=len(exon_children)>0)
        if lifton_gene is None or lifton_gene.ref_gene_id is None: return None            
    if len(exon_children) == 0:
        parent_feature = None
        if ENTRY_FEATURE: # Gene (1st) features without direct exons 
            parent_feature = lifton_gene
        else: # Middle features without exons
            parent_feature = lifton_gene.add_feature(copy.deepcopy(locus))
        features = tgt_db.children(locus, level=1)
        for feature in list(features):
            lifton_gene = evaluation(parent_feature, feature, ref_db, tgt_db, tree_dict, tgt_fai, ref_proteins, ref_trans, ref_features_dict, fw_score, args)
    else:
        if ENTRY_FEATURE: # Gene (1st) features with direct exons 
            ref_trans_id = ref_gene_id
        else: # Transcript features with direct exons 
            ref_gene_id, ref_trans_id = lifton_utils.get_ref_ids_liftoff(ref_features_dict, lifton_gene.entry.id, locus.id)
        lifton_status = lifton_class.Lifton_Status()
        lifton_trans, cds_num = lifton_add_trans_exon_cds_eval(lifton_gene, locus, ref_db, tgt_db, ref_trans_id, args)
        if lifton_trans == None: return lifton_gene
        lifton_trans_aln, lifton_aa_aln = lifton_gene.orf_search_protein(lifton_trans.entry.id, ref_trans_id, tgt_fai, ref_proteins, ref_trans, lifton_status, eval_only=True, eval_liftoff_chm13=args.evaluation_liftoff_chm13)
        lifton_utils.print_lifton_status(lifton_trans.entry.id, locus, lifton_status, DEBUG=args.debug)
        lifton_utils.write_lifton_eval_status(fw_score, lifton_trans.entry.id, locus, lifton_status)
    return lifton_gene


from lifton.liftoff import write_new_gff, liftover_types, polish, align_features, lift_features
import argparse
from pyfaidx import Fasta, Faidx

def main(arglist=None):
    args = parse_args(arglist)
    run_all_liftoff_steps(args)


def run_all_liftoff_steps(args, ref_db):
    if args.chroms is not None:
        ref_chroms, target_chroms = parse_chrm_files(args.chroms)
    else:
        ref_chroms = [args.reference]
        target_chroms = [args.target]
    parent_features_to_lift = get_parent_features_to_lift(args.features)
    lifted_feature_list = {}
    unmapped_features = []
    feature_db, feature_hierarchy, ref_parent_order = liftover_types.lift_original_annotation(ref_chroms, target_chroms,
                                                                                              lifted_feature_list, args,
                                                                                              unmapped_features,
                                                                                              parent_features_to_lift, ref_db)

    unmapped_features = map_unmapped_features(unmapped_features, target_chroms, lifted_feature_list, feature_db,
                                              feature_hierarchy, ref_parent_order, args)
    map_features_from_unplaced_seq(unmapped_features, lifted_feature_list, feature_db, feature_hierarchy,
                                   ref_parent_order, args)
    write_unmapped_features_file(args.u, unmapped_features)
    map_extra_copies(args, lifted_feature_list, feature_hierarchy, feature_db, ref_parent_order)

    if args.cds and args.polish is False:
        check_cds(lifted_feature_list, feature_hierarchy, args)
    if args.polish:
         print("polishing annotations")
         check_cds(lifted_feature_list, feature_hierarchy, args)
         write_new_gff.write_new_gff(lifted_feature_list, args, feature_db)
         find_and_polish_broken_cds(args, lifted_feature_list,feature_hierarchy, ref_chroms,
                                                          target_chroms,
                               unmapped_features, feature_db, ref_parent_order)
         if args.output != 'stdout':
             args.output += "_polished"
    write_new_gff.write_new_gff(lifted_feature_list, args, feature_db)



def parse_args(arglist):
    parser = argparse.ArgumentParser(description='Lift features from one genome assembly to another')
    parser.add_argument('target', help='target fasta genome to lift genes to')
    parser.add_argument('reference', help='reference fasta genome to lift genes from')

    refrgrp = parser.add_argument_group('Required input (annotation)')
    refrgrp.add_argument(
        '-g', '--reference-annotation', metavar='GFF',  required=True,
        help='the reference annotation file to lift over in GFF or GTF format (or) '
                'name of feature database; if not specified, the -g '
                'argument must be provided and a database will be built automatically'
    )

    # mxgrp = refrgrp.add_mutually_exclusive_group(required=True)
    # mxgrp.add_argument(
    #     '-g', metavar='GFF', help='annotation file to lift over in GFF or GTF format'
    # )
    # mxgrp.add_argument(
    #     '-db', metavar='DB', help='name of feature database; if not specified, the -g '
    #                               'argument must be provided and a database will be built automatically'
    # )

    outgrp = parser.add_argument_group('Output')
    outgrp.add_argument(
        '-o', '--output', default='stdout', metavar='FILE',
        help='write output to FILE in same format as input; by default, output is written to terminal (stdout)'
    )
    outgrp.add_argument(
        '-u', default='unmapped_features.txt', metavar='FILE',
        help='write unmapped features to FILE; default is "unmapped_features.txt"',
    )
    outgrp.add_argument(
        '-exclude_partial', action='store_true',
        help='write partial mappings below -s and -a threshold to unmapped_features.txt; if true '
             'partial/low sequence identity mappings will be included in the gff file with '
             'partial_mapping=True, low_identity=True in comments'
    )
    outgrp.add_argument(
        '-dir', "--directory", default='intermediate_files', metavar='DIR',
        help='name of directory to save intermediate fasta and SAM files; default is "intermediate_files"',
    )

    aligngrp = parser.add_argument_group('Alignments')
    aligngrp.add_argument('-mm2_options', metavar='=STR', type=str, default='-a --end-bonus '
                                                                            '5 --eqx -N 50 '
                                                                            '-p 0.5',
                          help='space delimited minimap2 parameters. By default ="-a --end-bonus 5 --eqx -N 50 -p 0.5"')
    aligngrp.add_argument(
        '-a', default=0.5, metavar='A', type=float,
        help='designate a feature mapped only if it aligns with coverage ≥A; by default A=0.5',
    )
    aligngrp.add_argument(
        '-s', default=0.5, metavar='S', type=float,
        help='designate a feature mapped only if its child features (usually exons/CDS) align '
             'with sequence identity ≥S; by default S=0.5'
    )
    aligngrp.add_argument(
        '-d', metavar='D', default=2.0, type=float,
        help='distance scaling factor; alignment nodes separated by more than a factor of D in '
             'the target genome will not be connected in the graph; by default D=2.0'
    )
    aligngrp.add_argument(
        '-flank', default=0, metavar='F', type=float, help="amount of flanking sequence to align as a "
                                                           "fraction [0.0-1.0] of gene length. This can improve gene "
                                                           "alignment where gene structure  differs between "
                                                           "target and "
                                                           "reference; by default F=0.0")

    parser.add_argument('-V', '--version', help='show program version', action='version', version='v1.6.3')
    parser.add_argument(
        '-p', '--threads', default=1, type=int, metavar='P', help='use p parallel processes to accelerate alignment; by default p=1'
    )
    parser.add_argument('-m', help='Minimap2 path', metavar='PATH')
    parser.add_argument('-f', '--features', metavar='TYPES', help='list of feature types to lift over')
    parser.add_argument(
        '-infer_genes', action='store_true', required=False,
        help='use if annotation file only includes transcripts, exon/CDS features'
    )
    parser.add_argument(
        '-infer_transcripts', action='store_true', required=False,
        help='use if annotation file only includes exon/CDS features and does not include transcripts/mRNA'
    )
    parser.add_argument(
        '-chroms', metavar='TXT', default=None, help='comma seperated file with corresponding chromosomes in '
                                       'the reference,target sequences',
    )
    parser.add_argument(
        '-unplaced', metavar='TXT',
        help='text file with name(s) of unplaced sequences to map genes from after genes from '
             'chromosomes in chroms.txt are mapped; default is "unplaced_seq_names.txt"',
    )
    parser.add_argument('-copies', action='store_true', help='look for extra gene copies in the target genome')
    parser.add_argument(
        '-sc', default=1.0, metavar='SC', type=float,
        help='with -copies, minimum sequence identity in exons/CDS for which a gene is considered '
             'a copy; must be greater than -s; default is 1.0',
    )
    parser.add_argument('-overlap', default=0.1, metavar='O', help="maximum fraction [0.0-1.0] of overlap allowed by 2 "
                                                                   "features; by default O=0.1", type=float)
    parser.add_argument('-mismatch', default=2, metavar='M', help="mismatch penalty in exons when finding best "
                                                                  "mapping; by default M=2", type=int)
    parser.add_argument('-gap_open', default=2, metavar='GO', help="gap open penalty in exons when finding best "
                                                                   "mapping; by default GO=2", type=int)
    parser.add_argument('-gap_extend', default=1, metavar='GE', help="gap extend penalty in exons when finding best "
                                                                     "mapping; by default GE=1", type=int)
    parser.add_argument('-subcommand', required=False,  help=argparse.SUPPRESS)
    parser.add_argument('-polish', required=False, action='store_true', default = False)
    parser.add_argument('-cds', required=False, action="store_true", default=True, help="annotate status of each CDS "
                                                                                        "(partial, missing start, "
                                                                                        "missing stop, inframe stop "
                                                                                        "codon)")
    parser._positionals.title = 'Required input (sequences)'
    parser._optionals.title = 'Miscellaneous settings'
    parser._action_groups = [parser._positionals, refrgrp, outgrp, aligngrp, parser._optionals]
    args = parser.parse_args(arglist)
    if '-a' not in args.mm2_options:
        args.mm2_options += ' -a'
    if '--eqx' not in args.mm2_options:
        args.mm2_options += ' --eqx'
    if '-N' not in args.mm2_options:
        args.mm2_options += " -N 50"
    if '-p' not in args.mm2_options:
        args.mm2_options += " -p 0.5"
    if '--end-bonus' not in args.mm2_options:
        args.mm2_options += "--end-bonus 5"
    if (float(args.s) > float(args.sc)):
        parser.error("-sc must be greater than or equal to -s")
    if (args.chroms is None and args.unplaced is not None):
        parser.error("-unplaced must be used with -chroms")
    return args


def parse_chrm_files(chroms_file):
    chroms = open(chroms_file, 'r')
    ref_chroms, target_chroms = [], []
    for line in chroms.readlines():
        ref_and_target_chrom = line.rstrip().split(",")
        ref_chroms.append(ref_and_target_chrom[0])
        if len(ref_and_target_chrom) > 1:
            target_chroms.append(ref_and_target_chrom[1])
    chroms.close()
    return ref_chroms, target_chroms


def get_parent_features_to_lift(feature_types_file):
    feature_types = ["gene"]
    if feature_types_file is not None:
        f = open(feature_types_file)
        for line in f.readlines():
            feature_types.append(line.rstrip())
    return feature_types


def map_unmapped_features(unmapped_features, target_chroms, lifted_feature_list, feature_db, feature_hierarchy,
                          ref_parent_order, args):
    if len(unmapped_features) > 0 and target_chroms[0] != args.target:
        print("mapping unaligned features to whole genome")
        ref_chroms = [args.reference]
        target_chroms = [args.target]
        return liftover_types.map_unmapped_genes_agaisnt_all(unmapped_features, ref_chroms, target_chroms,
                                                             lifted_feature_list, feature_db, feature_hierarchy,
                                                             ref_parent_order, args)
    return unmapped_features


def map_features_from_unplaced_seq(unmapped_features, lifted_feature_list, feature_db, feature_hierarchy,
                                   ref_parent_order, args):
    if args.unplaced is not None and args.chroms is not None:
        print("mapping unplaced genes")
        ref_chroms, target_chroms = parse_chrm_files(args.unplaced)
        target_chroms = [args.target]
        liftover_types.map_unplaced_genes(unmapped_features, ref_chroms, target_chroms,
                                          lifted_feature_list, feature_db, feature_hierarchy, ref_parent_order, args)


def write_unmapped_features_file(out_arg, unmapped_features):
    unmapped_out = open(out_arg, 'w')
    for gene in unmapped_features:
        unmapped_out.write(gene.id + "\n")
    unmapped_out.close()


def map_extra_copies(args, lifted_feature_list, feature_hierarchy, feature_db, ref_parent_order):
    if args.copies:
        print("mapping gene copies")
        ref_chroms = [args.reference]
        target_chroms = [args.target]
        liftover_types.map_extra_copies(ref_chroms, target_chroms, lifted_feature_list, feature_hierarchy, feature_db,
                                        ref_parent_order, args)


def find_and_polish_broken_cds(args, lifted_feature_list,feature_hierarchy, ref_chroms, target_chroms,
                               unmapped_features, feature_db, ref_parent_order,):
    args.subcommand = "polish"
    polish_lifted_features = {}
    ref_fa, target_fa = Fasta(args.reference), Fasta(args.target)
    for target_feature in lifted_feature_list:
        aligned_segments_new = {}
        if polish.polish_annotations(lifted_feature_list, ref_fa, target_fa, args, feature_hierarchy, target_feature):
            aligned_segments = align_features.align_features_to_target(ref_chroms, target_chroms, args,
                                                                       feature_hierarchy,
                                                                       "chrm_by_chrm", unmapped_features)
            aligned_segments_new[target_feature] = list(aligned_segments.values())[0]
            for seg in aligned_segments_new[target_feature]:
                seg.query_name = target_feature
            args.d = 100000000
            lift_features.lift_all_features(aligned_segments_new, args.a, feature_db, feature_hierarchy,
                                            unmapped_features, polish_lifted_features, args.s, None, args,
                                            ref_parent_order)

    check_cds(polish_lifted_features, feature_hierarchy, args)
    for feature in polish_lifted_features:
        original_feature = lifted_feature_list[feature][0]
        polished_feature = polish_lifted_features[feature][0]
        replace = False
        if 'valid_ORFs' not in polished_feature.attributes or int(polished_feature.attributes['valid_ORFs'][0]) > \
                int(original_feature.attributes['valid_ORFs'][0]):
            replace = True
        elif polished_feature.attributes['valid_ORFs'][0] == original_feature.attributes['valid_ORFs'][0]:
            if polished_feature.attributes['sequence_ID'][0] > original_feature.attributes['sequence_ID'][0]:
                replace = True
            elif polished_feature.attributes['coverage'][0] > original_feature.attributes['coverage'][0]:
                replace = True
        if replace:
            lifted_feature_list[feature] = polish_lifted_features[feature]


def check_cds(feature_list, feature_hierarchy, args):
    ref_faidx, target_faidx = Fasta(args.reference), Fasta(args.target)
    for target_feature in feature_list:
        target_sub_features = polish.get_sub_features(feature_list, target_feature)
        ref_sub_features = polish.get_sub_features(feature_hierarchy.children, target_sub_features[0].id)
        polish.find_and_check_cds(target_sub_features, ref_sub_features, ref_faidx,
                                                             target_faidx, feature_list[target_feature])



import subprocess
import os, copy
from lifton import align, lifton_class, logger, lifton_utils, protein_maximization, run_miniprot
from lifton.liftoff import liftoff_main
from lifton.liftoff.tests import test_basic, test_advanced
from intervaltree import Interval, IntervalTree

def run_liftoff(output_dir, ref_db, args):
    """
        This function runs liftoff.

        Parameters:
        - output_dir: output directory
        - args: Liftoff arguments

        Returns:
        liftoff_annotation: liftoff annotation file path
    """
    liftoff_args = copy.deepcopy(args)
    liftoff_outdir = output_dir + "liftoff/"    
    os.makedirs(liftoff_outdir, exist_ok=True)
    liftoff_annotation = liftoff_outdir + "liftoff.gff3"
    liftoff_args.output = liftoff_annotation
    liftoff_args.u = liftoff_outdir + "unmapped_features.txt"
    liftoff_main.run_all_liftoff_steps(liftoff_args, ref_db)
    if args.polish:
        liftoff_annotation += "_polished"
    # test_basic.test_yeast(liftoff_outdir + "test_basic/")
    # test_advanced.test_yeast(liftoff_outdir + "test_advance/")
    return liftoff_annotation


def initialize_lifton_gene(locus, ref_db, tree_dict, ref_features_dict, args, with_exons=False):
    """
        This function initializes Lifton gene instance.

        Parameters:
        - locus: gffutils feature instance
        - ref_db: reference database
        - tree_dict: intervaltree dictionary for each chromosome
        - ref_features_dict: reference features dictionary
        - with_exons: True if the gene has exons, False otherwise

        Returns:
        lifton_gene: Lifton gene instance
        ref_gene_id: reference gene
        ref_trans_id: reference transcript
    """
    ref_gene_id, ref_trans_id = lifton_utils.get_ref_ids_liftoff(ref_features_dict, locus.id, None)
    lifton_gene = lifton_class.Lifton_GENE(ref_gene_id, copy.deepcopy(locus), copy.deepcopy(ref_db[ref_gene_id].attributes), tree_dict, ref_features_dict, args, tmp=with_exons)
    return lifton_gene, ref_gene_id, ref_trans_id


def lifton_add_trans_exon_cds(lifton_gene, locus, ref_db, l_feature_db, ref_trans_id):
    """
        This function adds transcript, exons, and CDSs to the Lifton gene instance.

        Parameters:
        - lifton_gene: Lifton gene instance
        - locus: gffutils feature instance
        - ref_db: reference database
        - l_feature_db: liftoff feature database
        - ref_trans_id: reference transcript ID

        Returns:
        lifton_trans: Lifton transcript instance
        len(cdss_list): number of CDSs
    """
    lifton_trans = lifton_gene.add_transcript(ref_trans_id, copy.deepcopy(locus), copy.deepcopy(ref_db[ref_trans_id].attributes))
    exons = l_feature_db.children(locus, featuretype='exon', order_by='start')
    for exon in list(exons):
        lifton_gene.add_exon(lifton_trans.entry.id, exon)
    cdss = l_feature_db.children(locus, featuretype=('CDS', 'stop_codon'), order_by='start') 
    cdss_list = list(cdss)
    for cds in cdss_list:
        lifton_gene.add_cds(lifton_trans.entry.id, cds)
    return lifton_trans, len(cdss_list)


def process_liftoff_with_protein(locus, lifton_gene, lifton_trans,
                                 ref_id_2_m_id_trans_dict, m_feature_db, tree_dict,
                                 tgt_fai, ref_trans_id, ref_proteins, ref_trans,
                                 fw_chain, write_chains, lifton_status, DEBUG):
    """
        This function process liftoff annotation with protein.
        (1) Only Liftoff annotation: directly apply 
        (2) Liftoff annotation with miniprot annotation: Protein-maximization algorithm

        Parameters:
        - locus: gffutils feature instance
        - lifton_gene: Lifton gene instance
        - lifton_trans: Lifton transcript instance
        - ref_id_2_m_id_trans_dict: reference id to miniprot transcript ids dictionary
        - m_feature_db: miniprot feature database
        - tree_dict: intervaltree dictionary for each chromosome
        - tgt_fai: target fasta index
        - ref_trans_id: reference transcript ID
        - ref_proteins: reference proteins dictionary
        - ref_trans: reference transcript dictionary
        - fw_chain: file writer for chains
        - write_chains: write chains or not
        - lifton_status: Lifton_Status instance
        - DEBUG: debug mode
    """
    # Liftoff alignment
    liftoff_aln = lifton_utils.LiftOn_liftoff_alignment(lifton_trans, locus, tgt_fai, ref_proteins, ref_trans_id, lifton_status)
    # miniprot alignment
    miniprot_aln, has_valid_miniprot = lifton_utils.LiftOn_miniprot_alignment(locus.seqid, locus, ref_id_2_m_id_trans_dict, m_feature_db, tree_dict, tgt_fai, ref_proteins, ref_trans_id, lifton_status)
    if liftoff_aln is None:
        lifton_status.annotation = "no_ref_protein"
    elif liftoff_aln.identity == 1:
        # Liftoff protein annotation is perfect"
        lifton_status.lifton_aa = 1
    elif liftoff_aln.identity < 1:
        # Liftoff protein annotation is not perfect
        if has_valid_miniprot:
            lifton_status.annotation = "LiftOn_chaining_algorithm"
            cds_list, chains = protein_maximization.chaining_algorithm(liftoff_aln, miniprot_aln, tgt_fai, DEBUG)
            if write_chains:
                lifton_utils.write_lifton_chains(fw_chain, lifton_trans.entry.id, chains)
            lifton_gene.update_cds_list(lifton_trans.entry.id, cds_list)


def process_liftoff(lifton_gene, locus, ref_db, l_feature_db, ref_id_2_m_id_trans_dict, m_feature_db, tree_dict, tgt_fai, ref_proteins, ref_trans, ref_features_dict, fw_score, fw_chain, args, ENTRY_FEATURE=False):
    """
        This function processes liftoff annotation.

        Parameters:
        - lifton_gene: Lifton gene instance
        - locus: feature instance 
        - ref_db: reference database
        - l_feature_db: liftoff feature database
        - ref_id_2_m_id_trans_dict: reference id to miniprot transcript ids dictionary
        - m_feature_db: miniprot feature database
        - tree_dict: intervaltree dictionary for each chromosome
        - tgt_fai: target fasta index
        - ref_proteins: reference protein dictionary
        - ref_trans: reference transcript dictionary
        - ref_features_dict: reference features dictionary
        - fw_score: file writer for scores
        - fw_chain: file writer for chains
        - write_chains: write chains or not
        - DEBUG: debug mode
        - ENTRY_FEATURE: True if the feature is the root feature for a gene locus

        Returns:
        lifton_gene: LiftOn gene instance
    """
    exon_children = list(l_feature_db.children(locus, featuretype='exon', level=1, order_by='start'))
    if lifton_gene is None and ENTRY_FEATURE:   
        # Gene (1st) features
        lifton_gene, ref_gene_id, ref_trans_id = initialize_lifton_gene(locus, ref_db, tree_dict, ref_features_dict, args, with_exons=len(exon_children)>0)
        if lifton_gene.ref_gene_id is None: return None            
    if len(exon_children) == 0:
        parent_feature = None
        if ENTRY_FEATURE: # Gene (1st) features without direct exons 
            parent_feature = lifton_gene
        else: # Middle features without exons
            parent_feature = lifton_gene.add_feature(copy.deepcopy(locus))
        features = l_feature_db.children(locus, level=1)
        for feature in list(features):
            lifton_gene = process_liftoff(parent_feature, feature, ref_db, l_feature_db, ref_id_2_m_id_trans_dict, m_feature_db, tree_dict, tgt_fai, ref_proteins, ref_trans, ref_features_dict, fw_score, fw_chain, args)
    else:
        if ENTRY_FEATURE: # Gene (1st) features with direct exons 
            ref_trans_id = ref_gene_id
        else: # Transcript features with direct exons 
            ref_gene_id, ref_trans_id = lifton_utils.get_ref_ids_liftoff(ref_features_dict, lifton_gene.entry.id, locus.id)
        lifton_status = lifton_class.Lifton_Status()
        lifton_status.annotation = "Liftoff"
        try: # Test if the reference transcript exists. Skip if not.
            ref_db[ref_trans_id]
        except:
            return None
        lifton_trans, cds_num = lifton_add_trans_exon_cds(lifton_gene, locus, ref_db, l_feature_db, ref_trans_id)
        if cds_num > 0:
            process_liftoff_with_protein(locus, lifton_gene, lifton_trans,
                                        ref_id_2_m_id_trans_dict, m_feature_db, tree_dict,
                                        tgt_fai, ref_trans_id, ref_proteins, ref_trans,
                                        fw_chain, args.write_chains, lifton_status, args.debug)
        if not args.no_orf_search:
            lifton_trans_aln, lifton_aa_aln = lifton_gene.orf_search_protein(lifton_trans.entry.id, ref_trans_id, tgt_fai, ref_proteins, ref_trans, lifton_status) 
        # lifton_utils.print_lifton_status(lifton_trans.entry.id, locus, lifton_status, DEBUG=args.debug)
        lifton_gene.add_lifton_gene_status_attrs("Liftoff")
        lifton_gene.add_lifton_trans_status_attrs(lifton_trans.entry.id, lifton_status)
        lifton_utils.write_lifton_status(fw_score, lifton_trans.entry.id, locus, lifton_status)
    return lifton_gene


import parasail
from Bio.Seq import Seq
from cigar import Cigar
from lifton import get_id_fraction, lifton_class


def adjust_cdss_protein_boundary(cdss_protein_aln_boundary, cigar_accum_len, length):
    """
        This function adjust CDS protein boundaries based on CIGAR string on protein alignment.

        Parameters:
        - cdss_protein_aln_boundary: CDS protein alignment boundary
        - cigar_accum_len: accumulated length of CIGAR string
        - length: length of CIGAR string

        Returns:
        cdss_protein_aln_boundary: adjusted CDS protein alignment boundary
    """
    cds_boundary_shift = 0
    for i in range(len(cdss_protein_aln_boundary)):
        cdss_start = cdss_protein_aln_boundary[i][0] + cds_boundary_shift
        cdss_end = cdss_protein_aln_boundary[i][1] + cds_boundary_shift
        if (cdss_start <= cigar_accum_len) and (cdss_end >= cigar_accum_len):
            cds_boundary_shift += length
            cdss_end += length
        cdss_protein_aln_boundary[i] = (cdss_start, cdss_end)
    return cdss_protein_aln_boundary


def get_cdss_protein_boundary(cdss_lens):
    """
        This function maps the CDSs boundaries on to the protein.

        Parameters:
        - cdss_lens: list of CDS lengths

        Returns:
        cdss_protein_boundary: CDS protein boundary
    """
    cdss_cumulative = [sum(cdss_lens[:i+1]) for i in range(len(cdss_lens))]
    cdss_cumulative_div = [x / 3 for x in cdss_cumulative]
    cdss_protein_boundary = {}
    for idx in range(len(cdss_cumulative_div)):
        start = cdss_cumulative_div[idx-1] if idx > 0 else 0
        end = cdss_cumulative_div[idx]
        cdss_protein_boundary[idx] = (start, end)
    return cdss_protein_boundary


def parasail_align_protein_base(protein_seq, ref_protein_seq):
    """
        This is the base function that uses parasail to align the protein sequence to the reference protein sequence.

        Parameters:
        - protein_seq: protein sequence in string format
        - ref_protein_seq: reference protein sequence in string format

        Returns:
        extracted_parasail_res: parasail alignment result
    """
    matrix = parasail.Matrix("blosum62")
    gap_open = 11
    gap_extend = 1
    protein_seq = "*" if protein_seq == "" else protein_seq
    # Return: (Query, Target)
    extracted_parasail_res = parasail.nw_trace_scan_sat(protein_seq, ref_protein_seq, gap_open, gap_extend, matrix)
    return extracted_parasail_res


def protein_align(protein_seq, ref_protein_seq):
    """
        This function aligns the protein sequence to the reference protein sequence and extracts the alignment information.

        Parameters:
        - protein_seq: protein sequence in string format
        - ref_protein_seq: reference protein sequence in string format

        Returns:
        lifton_aln: Lifton_Alignment object
    """
    extracted_parasail_res = parasail_align_protein_base(protein_seq, ref_protein_seq)
    alignment_query = extracted_parasail_res.traceback.query
    alignment_comp = extracted_parasail_res.traceback.comp
    alignment_ref = extracted_parasail_res.traceback.ref
    extracted_matches, extracted_length = get_id_fraction.get_AA_id_fraction(extracted_parasail_res.traceback.ref, extracted_parasail_res.traceback.query)
    extracted_identity = extracted_matches/extracted_length
    lifton_aln = lifton_class.Lifton_Alignment(extracted_identity, None, alignment_query, alignment_comp, alignment_ref, None, None, protein_seq, ref_protein_seq, None)
    return lifton_aln


def parasail_align_DNA_base(trans_seq, ref_trans_seq):
    """
        This is the base function that uses parasail to align the transcript sequence to the reference transcript sequence.

        Parameters:
        - trans_seq: transcript sequence in string format
        - ref_trans_seq: reference transcript sequence in string format

        Returns:
        extracted_parasail_res: parasail alignment result
    """
    matrix = parasail.matrix_create("ACGT*", 1, -3)
    gap_open = 5
    gap_extend = 2
    # Return: (Query, Target)
    extracted_parasail_res = parasail.nw_trace_scan_sat(trans_seq, ref_trans_seq, gap_open, gap_extend, matrix)
    return extracted_parasail_res


def trans_align(trans_seq, ref_trans_seq):
    """
        This function aligns the transcript sequence to the reference transcript sequence and extracts the alignment information.

        Parameters:
        - trans_seq: transcript sequence in string format
        - ref_trans_seq: reference transcript sequence in string format

        Returns:
        lifton_aln: Lifton_Alignment object
    """
    extracted_parasail_res = parasail_align_DNA_base(trans_seq, ref_trans_seq)
    alignment_query = extracted_parasail_res.traceback.query
    alignment_comp = extracted_parasail_res.traceback.comp
    alignment_ref = extracted_parasail_res.traceback.ref
    # print("alignment_query: ", alignment_query) 
    # print("alignment_comp: ", alignment_comp) 
    # print("alignment_ref: ", alignment_ref) 
    extracted_matches, extracted_length = get_id_fraction.get_DNA_id_fraction(extracted_parasail_res.traceback.ref, extracted_parasail_res.traceback.query)
    extracted_identity = extracted_matches/extracted_length
    lifton_aln = lifton_class.Lifton_Alignment(extracted_identity, None, alignment_query, alignment_comp, alignment_ref, None, None, trans_seq, ref_trans_seq, None)
    return lifton_aln


def LiftOn_translate(lifton_trans, fai, ref_proteins, ref_trans_id):
    """
        This function translates the given annotation.
        
        Parameters:
        - lifton_trans: LiftOn transcript instance
        - fai: fasta index
        - ref_proteins: reference protein dictionary
        - ref_trans_id: reference transcript ID

        Returns:
        Referece protein sequence (in string), translated protein sequence (in string), CDS lengths, CDS children
    """
    coding_seq, cds_children, cdss_lens = lifton_trans.get_coding_seq(fai)
    coding_seq, _ = lifton_trans.get_coding_trans_seq(fai)
    protein_seq = lifton_trans.translate_coding_seq(coding_seq)
    # print("protein_seq: ", protein_seq)
    return str(ref_proteins[ref_trans_id]), protein_seq, cdss_lens, cds_children


def lifton_parasail_align(lifton_trans, db_entry, fai, ref_proteins, ref_trans_id):
    """
        This function aligns the annotated protein sequence to the reference protein sequence

        Parameters:
        - lifton_trans: LiftOn transcript instance
        - db_entry: database entry
        - fai: fasta index
        - ref_proteins: reference protein dictionary
        - ref_trans_id: reference transcript ID
        - lifton_status: LiftOn status

        Returns:
        aln: Lifton_Alignment object
    """
    # Step 1: Check if the ref_trans_id is in the reference protein dictionary
    aln = None
    if ref_trans_id not in ref_proteins.keys():
        return aln
    # Step 2: Translate the given annotation
    ref_protein_seq, protein_seq, cdss_lens, cds_children = LiftOn_translate(lifton_trans, fai, ref_proteins, ref_trans_id)
    # Step 3: Get the corresponding protein boundaries for each CDS
    cdss_protein_boundary = get_cdss_protein_boundary(cdss_lens)
    # Step 4: Protein alignment
    if protein_seq == None:
        return aln
    extracted_parasail_res = parasail_align_protein_base(protein_seq, ref_protein_seq)
    # Step 5: Extract the alignment information
    alignment_query = extracted_parasail_res.traceback.query
    alignment_comp = extracted_parasail_res.traceback.comp
    alignment_ref = extracted_parasail_res.traceback.ref
    cigar = extracted_parasail_res.cigar
    decoded_cigar = cigar.decode.decode()
    cigar_ls = list(Cigar(decoded_cigar).items())
    # Step 6: Change the CDS protein boundaries based on CIGAR string.
    cigar_accum_len = 0
    cdss_protein_aln_boundary = cdss_protein_boundary.copy()
    for length, symbol in cigar_ls:
        if symbol == "D":
            # print(length, symbol)
            cdss_protein_aln_boundary = adjust_cdss_protein_boundary(cdss_protein_aln_boundary, cigar_accum_len, length)
        cigar_accum_len += length
        # print("cigar_accum_len: ", cigar_accum_len)
    extracted_matches, extracted_length = get_id_fraction.get_AA_id_fraction(extracted_parasail_res.traceback.ref, extracted_parasail_res.traceback.query)
    extracted_identity = extracted_matches/extracted_length
    aln = lifton_class.Lifton_Alignment(extracted_identity, cds_children, alignment_query, alignment_comp, alignment_ref, cdss_protein_boundary, cdss_protein_aln_boundary, protein_seq, ref_protein_seq, db_entry)
    return aln



import gffutils
from pyfaidx import Fasta, Faidx
from lifton.liftoff  import liftoff_utils, feature_hierarchy, new_feature
import os
import sys
import numpy as np
import ujson as json





def extract_features_to_lift(ref_chroms, liftover_type, parents_to_lift, ref_db, args):
    print("extracting features")
    if os.path.exists(args.directory) is False:
        os.mkdir(args.directory)
    # feature_db = create_feature_db_connections(args)
    feature_db = ref_db.db_connection
    feature_hierarchy, parent_order = seperate_parents_and_children(feature_db, parents_to_lift)
    get_gene_sequences(feature_hierarchy.parents, ref_chroms, args, liftover_type)
    return feature_hierarchy, feature_db, parent_order


# def create_feature_db_connections(args):
#     gffutils.constants.ignore_url_escape_characters = True
#     if args.infer_transcripts is True:
#         disable_transcripts = False
#     else:
#         disable_transcripts = True
#     if args.infer_genes is True:
#         disable_genes = False
#     else:
#         disable_genes = True
#     feature_db = build_database(args.db, args.g, disable_transcripts, disable_genes)
#     return feature_db



# def build_database(db, gff_file, disable_transcripts, disable_genes,):
#     if db is None:
#         try:
#             feature_db = gffutils.create_db(gff_file, gff_file + "_db", merge_strategy="create_unique", force=True,
#                                             disable_infer_transcripts=disable_transcripts,
#                                             disable_infer_genes=disable_genes, verbose=True)
#         except:
#             find_problem_line(gff_file)
#     else:
#         feature_db = gffutils.FeatureDB(db)
#     return feature_db



def create_feature_db_connections(args):
    try:
        feature_db = gffutils.FeatureDB(args.reference_annotation)
    except:
        feature_db = build_database(args)
    feature_db.execute('ANALYZE features')
    return feature_db


    # ref_db = annotation.Annotation(args.reference_annotation, args.infer_genes, args.infer_transcripts, args.merge_strategy, args.id_spec, args.force, args.verbose)


def build_database(args):
    if args.infer_genes:
        disable_genes = False
    else:
        disable_genes = True
    try:
        transform_func = get_transform_func(args)
        feature_db = gffutils.create_db(args.reference_annotation, args.reference_annotation + "_db", merge_strategy="create_unique",
                                    force=True,
                                    verbose=True, disable_infer_transcripts=True,
                                        disable_infer_genes=disable_genes, transform=transform_func)
    except Exception as e:
        print("gffutils database build failed with", e)
        sys.exit()
    return feature_db

def get_transform_func(args):
    if args.infer_genes is False:
        return None
    else:
        return transform_func

def transform_func(x):
    if 'transcript_id' in x.attributes:
        x.attributes['transcript_id'][0] += '_transcript'
    return x






def find_problem_line(gff_file):
    f = open(gff_file, 'r')
    lines = f.readlines()
    for i in range (len(lines)):
        line = lines[i]
        if line[0] != "#":
            try:
                gffutils.create_db(line, ":memory:", from_string=True, force=True)
            except:
                sys.exit("ERROR:Incorrect GFF/GTF syntax on line " + str(i + 1))


def seperate_parents_and_children(feature_db, parent_types_to_lift):
    c = feature_db.conn.cursor()
    relations = [list(feature) for feature in c.execute('''SELECT * FROM relations join features as a on 
    a.id = relations.parent join features as b on b.id = relations.child''') if
                 feature[0] != feature[1]]
    all_ids = [list(feature)[0]for feature in c.execute('''SELECT * FROM features''')]
    all_children_ids = [relation[1] for relation in relations ]
    all_parent_ids = [relation[0] for relation in relations]
    lowest_children = np.setdiff1d(all_ids,  all_parent_ids )
    highest_parents = np.setdiff1d(all_ids, all_children_ids)
    intermediates = set(all_children_ids).intersection(set( all_parent_ids ))
    parent_dict, child_dict, intermediate_dict = {}, {}, {}
    add_parents(parent_dict, child_dict, highest_parents, parent_types_to_lift, feature_db)
    add_children(parent_dict, child_dict, lowest_children, feature_db)
    add_intermediates(intermediates, intermediate_dict, feature_db)
    parent_order = liftoff_utils.find_parent_order(
        [parent for parent in list(parent_dict.values()) if parent is not None])
    ref_feature_hierarchy = feature_hierarchy.feature_hierarchy(parent_dict, intermediate_dict, child_dict)
    return ref_feature_hierarchy, parent_order


def add_parents(parent_dict, child_dict, highest_parents, parent_types_to_lift, feature_db):
    c = feature_db.conn.cursor()
    cond = ', '.join('"{0}"'.format(w) for w in highest_parents)
    query =  "SELECT * FROM features WHERE id IN ({})".format(cond)
    for result in c.execute(query):
        feature_tup = tuple(result)
        parent = new_feature.new_feature(feature_tup[0], feature_tup[3], feature_tup[1], feature_tup[2],feature_tup[7],
                                          feature_tup[4], feature_tup[5], feature_tup[8], json.loads(feature_tup[9]))
        if parent.featuretype in parent_types_to_lift:
            parent_dict[parent.id] = parent
            child_dict[parent.id] = []


def add_children(parent_dict, child_dict, lowest_children, feature_db):
    c = feature_db.conn.cursor()
    cond = ', '.join('"{0}"'.format(w) for w in lowest_children)
    query = "select * from relations join features on features.id  = relations.child where relations.child IN ({})".format(cond)
    c.execute(query)
    results = c.fetchall()
    added_children_ids = []
    for result in results:
        feature_tup = tuple(result)
        parent = feature_tup[0]
        if parent in parent_dict:
            child = new_feature.new_feature(feature_tup[3], feature_tup[6], feature_tup[4], feature_tup[5],
                                            feature_tup[10],
                                          feature_tup[7], feature_tup[8], feature_tup[11], json.loads(feature_tup[12]))
            if child.featuretype != "intron":
                if "Parent" not in child.attributes:
                    add_parent_tag(child, feature_db)
                child_dict[parent].append(child)
                added_children_ids.append(child.id)
    single_level_features = np.setdiff1d(lowest_children, added_children_ids)
    for feature in single_level_features:
        if feature in parent_dict:
            child_dict[feature] = [parent_dict[feature]]




def add_parent_tag(feature, feature_db):
    parent_id = ""
    parents = [parent for parent in feature_db.parents(feature.id, level=1) if feature.id != parent.id]
    if len(parents) > 0:
        parent_id = parents[0].id
    else:
        parents = [parent for parent in feature_db.parents(feature.id) if feature.id != parent.id]
        if len(parents) > 0:
            parent_id = parents[0].id
    feature.attributes["Parent"] = [parent_id]



def add_intermediates(intermediate_ids, intermediate_dict, feature_db):
    c = feature_db.conn.cursor()
    cond = ', '.join('"{0}"'.format(w) for w in intermediate_ids)
    query =  "select * from features where id IN ({})".format(cond)
    for result in c.execute(query):
        feature_tup = tuple(result)
        intermediate_feature = new_feature.new_feature(feature_tup[0], feature_tup[3], feature_tup[1], feature_tup[2],
                                           feature_tup[7],
                                          feature_tup[4], feature_tup[5], feature_tup[8], json.loads(feature_tup[9]))
        intermediate_dict[intermediate_feature.id] = intermediate_feature
        if "Parent" not in intermediate_feature.attributes:
            add_parent_tag(intermediate_feature, feature_db)


def get_gene_sequences(parent_dict, ref_chroms, args, liftover_type):
    fai = Fasta(args.reference)
    if liftover_type == "unplaced":
        open(args.directory + "/unplaced_genes.fa", 'w')
    for chrom in ref_chroms:
        fasta_out = get_fasta_out(chrom, args.reference, liftover_type, args.directory)
        sorted_parents = sorted(list(parent_dict.values()), key=lambda x: x.seqid)

        if len(sorted_parents) == 0:
            sys.exit(
                "GFF does not contain any gene features. Use -f to provide a list of other feature types to lift over.")
        write_gene_sequences_to_file(chrom, args.reference, fai, sorted_parents, fasta_out, args)
        fasta_out.close()


def get_fasta_out(chrom_name, reference_fasta_name, liftover_type, inter_files):
    if chrom_name == reference_fasta_name and (liftover_type == "chrm_by_chrm" or liftover_type == "copies"):
        fasta_out_name = "reference_all"
    elif liftover_type == "unmapped":
        fasta_out_name = "unmapped_to_expected_chrom"
    elif liftover_type == "unplaced":
        fasta_out_name = "unplaced"
    else:
        fasta_out_name = chrom_name
    if liftover_type == "unplaced":
        fasta_out = open(inter_files + "/" + fasta_out_name + "_genes.fa", 'a')
    else:
        fasta_out = open(inter_files + "/" + fasta_out_name + "_genes.fa", 'w')
    return fasta_out


def write_gene_sequences_to_file(chrom_name, reference_fasta_name, reference_fasta_idx, parents, fasta_out, args):
    if chrom_name == reference_fasta_name:
        current_chrom = parents[0].seqid
    else:
        current_chrom = chrom_name
    chrom_seq = reference_fasta_idx[current_chrom][:].seq
    for parent in parents:
        # Skip those chromosomes that are not in the reference fasta
        if parent.seqid not in reference_fasta_idx.keys():
            continue

        if parent.seqid != current_chrom and chrom_name == reference_fasta_name:
            current_chrom = parent.seqid
            chrom_seq = reference_fasta_idx[current_chrom][:].seq
        if parent.seqid == chrom_name or chrom_name == reference_fasta_name:
            gene_length = parent.end - parent.start + 1
            parent.start = round(max(1, parent.start - args.flank * gene_length))
            parent.end = round(min(parent.end + args.flank * gene_length, len(chrom_seq)))
            parent_seq = chrom_seq[parent.start - 1: parent.end]
            fasta_out.write(">" + parent.id + "\n" + str(parent_seq) + "\n")



from lifton.liftoff  import lift_features, liftoff_utils
import numpy as np
from interlap import InterLap


def fix_incorrectly_overlapping_features(all_lifted_features, features_to_check, all_aligned_segs,
                                         unmapped_features, threshold, feature_hierarchy, feature_db,
                                         ref_parent_order, seq_id_threshold, distance_factor,
                                         max_overlap):
    features_to_remap, feature_locations = check_homologues(all_lifted_features, features_to_check,
                                                            feature_hierarchy.parents,
                                                            ref_parent_order, max_overlap)
    resolve_overlapping_homologues(all_aligned_segs, all_lifted_features, features_to_remap, unmapped_features,
                                   threshold, feature_hierarchy, feature_db, ref_parent_order, seq_id_threshold,
                                   feature_locations, distance_factor, max_overlap)


def check_homologues(all_lifted_features, lifted_features_to_check, parent_dict, ref_parent_order, max_overlap):
    all_feature_list = liftoff_utils.get_parent_list(all_lifted_features)
    features_to_check_list = liftoff_utils.get_parent_list(lifted_features_to_check)
    target_parent_order = liftoff_utils.find_parent_order(all_feature_list)
    remap_features = set()
    feature_locations = build_interval_list(all_feature_list)
    for feature in features_to_check_list:
        overlaps = liftoff_utils.find_overlaps(feature.start - 1, feature.end - 1, feature.seqid, feature.strand,
                                               feature.attributes["copy_num_ID"][0], feature_locations, parent_dict,
                                               all_lifted_features, max_overlap)
        for overlap in overlaps:
            if overlap[2][0] != feature.attributes["copy_num_ID"][0]:
                feature_to_compare = overlap[2][1]
                compare_overlapping_feature(feature_to_compare, feature, remap_features, ref_parent_order,
                                            target_parent_order)
    return remap_features, feature_locations


def build_interval_list(features):
    inter = InterLap()
    feature_coords = [[feature.start - 1, feature.end - 1, [feature.attributes["copy_num_ID"][0], feature]] for feature in
                      features]
    if len(feature_coords) > 0:
        inter.update(feature_coords)
    return inter


def compare_overlapping_feature(overlapping_feature, feature, remap_features, ref_parent_order,
                                target_parent_order):
    feature_to_remap = find_feature_to_remap(feature, overlapping_feature,
                                             ref_parent_order, target_parent_order, remap_features)
    remap_features.add(feature_to_remap.attributes["copy_num_ID"][0])


def find_feature_to_remap(feature, overlap_feature, ref_parent_order, target_parent_order, remap_features):
    feature_is_copy = is_copy(feature)
    overlap_feature_is_copy = is_copy(overlap_feature)
    if feature_is_copy and overlap_feature_is_copy is False:
        return feature
    if overlap_feature_is_copy and feature_is_copy is False:
        return overlap_feature
    if already_in_list(feature, remap_features):
        return feature
    if already_in_list(overlap_feature, remap_features):
        return overlap_feature
    if has_greater_seq_id(feature, overlap_feature):
        return overlap_feature
    if has_greater_seq_id(overlap_feature, feature):
        return feature
    farthest_neighbor_distance = find_out_of_order_feature(feature_is_copy, ref_parent_order,
                                                           target_parent_order, feature,
                                                           overlap_feature_is_copy, overlap_feature)
    if farthest_neighbor_distance is not None:
        return farthest_neighbor_distance
    if is_shorter(feature, overlap_feature):
        return feature
    if is_shorter(overlap_feature, feature):
        return overlap_feature
    return overlap_feature


def is_copy(feature):
    return feature.attributes["copy_num_ID"][0][-2:] != '_0'


def already_in_list(feature, remap_features):
    return feature.attributes["copy_num_ID"][0] in remap_features


def has_greater_seq_id(feature1, feature2):
    if feature1.score < feature2.score:
        return True
    return False


def find_out_of_order_feature(feature_is_copy, ref_parent_order, target_parent_order, feature,
                              overlap_feature_is_copy, overlap_feature):
    if feature_is_copy is False:
        neighbor_distance_feature = check_order(ref_parent_order, target_parent_order, feature, overlap_feature)
    else:
        neighbor_distance_feature = None
    if overlap_feature_is_copy is False:
        neighbor_distance_overlap_feature = check_order(ref_parent_order, target_parent_order, overlap_feature, feature)
    else:
        neighbor_distance_overlap_feature = None
    if neighbor_distance_feature is not None and neighbor_distance_overlap_feature is not None:
        if neighbor_distance_feature < neighbor_distance_overlap_feature:
            return overlap_feature
        else:
            return feature
    return None


def check_order(ref_parent_order, target_parent_order, feature, overlap_feature):
    ref_neighbor_upstream = liftoff_utils.find_nonoverlapping_upstream_neighbor(ref_parent_order, feature.id)
    if ref_neighbor_upstream is not None:
        ref_distance = find_distance_between_features(ref_neighbor_upstream, target_parent_order, feature,
                                                      overlap_feature)
        return ref_distance
    return None


def find_distance_between_features(ref_neighbor_upstream, target_parent_order, feature, overlap_feature):
    feature_location_arr = np.where(target_parent_order[:, 0] == feature.id)[0]
    neighbor_location_arr = np.where(target_parent_order[:, 0] == ref_neighbor_upstream)[0]
    if len(feature_location_arr) > 0 and len(neighbor_location_arr) > 0:
        feature_location, neighbor_location = feature_location_arr[0], neighbor_location_arr[0]
        features_in_range = target_parent_order[min(feature_location, neighbor_location):max(feature_location,
                                                                                             neighbor_location) + 1]
        distance = np.abs(feature_location - neighbor_location)
        if overlap_feature.id in features_in_range[:, 0]:
            distance -= 1
        return distance
    return None


def is_shorter(feature1, feature2):
    return (feature1.end - feature1.start) < (feature2.end - feature2.start)


def resolve_overlapping_homologues(all_aligned_segs, lifted_feature_list, features_to_remap, unmapped_features,
                                   threshold, feature_hierarchy, feature_db, ref_parent_order, seq_id_threshold,
                                   feature_locations, distance_factor, max_overlap):
    iter = 0
    max_iter = 10 * len(features_to_remap)
    while len(features_to_remap) > 0:
        iter += 1
        if iter > max_iter:
            break
        aligned_segs_for_remap = remove_features_and_get_alignments(features_to_remap, lifted_feature_list,
                                                                    all_aligned_segs)
        lift_features.lift_all_features(aligned_segs_for_remap, threshold, feature_db, feature_hierarchy,
                                        unmapped_features,
                                        lifted_feature_list, seq_id_threshold, feature_locations, distance_factor,
                                        ref_parent_order)
        features_to_check = get_successfully_remapped_features(lifted_feature_list, features_to_remap,
                                                               )
        features_to_remap, feature_locations = check_homologues(lifted_feature_list, features_to_check,
                                                                feature_hierarchy.parents,
                                                                ref_parent_order, max_overlap)
    remove_unresolved_features(features_to_remap, feature_hierarchy.parents, lifted_feature_list, unmapped_features)
    return lifted_feature_list


def remove_features_and_get_alignments(features_to_remap, lifted_feature_list, all_aligned_segs):
    aligned_segs_for_remap = {}
    for feature_to_remap in features_to_remap:
        del lifted_feature_list[feature_to_remap]
        aligned_segs_for_remap[feature_to_remap] = all_aligned_segs[feature_to_remap]
    return aligned_segs_for_remap


def get_successfully_remapped_features(lifted_feature_list, features_to_remap):
    successfully_remapped_features = {}
    for feature_to_remap in features_to_remap:
        if feature_to_remap in lifted_feature_list:
            successfully_remapped_features[feature_to_remap] = lifted_feature_list[feature_to_remap]
    return successfully_remapped_features


def remove_unresolved_features(features_to_remap, parent_dict, lifted_feature_list, unmapped_features):
    for feature in features_to_remap:
        unmapped_features.append(parent_dict[liftoff_utils.convert_id_to_original(feature)])
        del lifted_feature_list[feature]



from lifton.liftoff  import liftoff_utils
from lifton import __version__
import sys


def write_header(f, out_type):
    if out_type == 'gff3':
        f.write('##gff-version 3' + "\n")
    f.write("# LiftOn v" + __version__ + "\n")
    f.write("# " + " ".join(sys.argv) + "\n")


def write_new_gff(lifted_features, args, feature_db):
    if args.output != 'stdout':
        f = open(args.output, 'w')
    else:
        f = sys.stdout
    out_type = feature_db.dialect['fmt']
    write_header(f, out_type)
    parents = liftoff_utils.get_parent_list(lifted_features)
    parents.sort(key=lambda x: x.id)
    final_parent_list = finalize_parent_features(parents, args)
    final_parent_list.sort(key=lambda x: (x.seqid, x.start))
    for final_parent in final_parent_list:
        child_features = lifted_features[final_parent.attributes["copy_id"][0]]
        parent_child_dict = build_parent_dict(child_features, final_parent)
        write_feature([final_parent], f, child_features, parent_child_dict, out_type)


def finalize_parent_features(parents, args):
    final_parent_list = []
    copy_num_dict = {}
    for parent in parents:
        add_to_copy_num_dict(parent, copy_num_dict)
        copy_num = copy_num_dict[parent.id]
        add_attributes(parent, copy_num, args)
        final_parent_list.append(parent)
    return final_parent_list


def add_to_copy_num_dict(parent, copy_num_dict):
    if parent.id in copy_num_dict:
        copy_num_dict[parent.id] += 1
    else:
        copy_num_dict[parent.id] = 0


def add_attributes(parent, copy_num, args):
    parent.score = "."
    keys_to_readd_at_end = ["copy_num_ID", "partial_mapping", "low_identity", "extra_copy_number",
                            "partial_mapping",  "low_identity"]
    if "copy_id" not in parent.attributes:
        parent.attributes["copy_id"] = parent.attributes["copy_num_ID"]
    for key in keys_to_readd_at_end:
        if key in parent.attributes:
            del  parent.attributes[key]
    parent.attributes["extra_copy_number"] = [str(copy_num)]
    parent.attributes["copy_num_ID"] = [parent.id + "_" + str(copy_num)]

    if float(parent.attributes["coverage"][0]) < args.a:
        parent.attributes["partial_mapping"] = ["True"]
    if float(parent.attributes["sequence_ID"][0]) < args.s:
        parent.attributes["low_identity"] = ["True"]


def build_parent_dict(child_features, final_parent):
    parent_child_dict = {}
    for child in child_features:
        if "Parent" in child.attributes:
            child.attributes["extra_copy_number"] = final_parent.attributes["extra_copy_number"]
            if child.attributes["Parent"][0] in parent_child_dict:
                parent_child_dict[child.attributes["Parent"][0]].append(child)
            else:
                parent_child_dict[child.attributes["Parent"][0]] = [child]
    return parent_child_dict


def write_feature(children, outfile, child_features, parent_dict, output_type):
    for child in children:
        write_line(child, outfile, output_type)
        if child.id in parent_dict:
            new_children = parent_dict[child.id ]
            write_feature(new_children, outfile, child_features, parent_dict, output_type)
    return


def write_line(feature, out_file, output_type):
    if feature.attributes["extra_copy_number"][0]!='0':
        attr_dict = edit_copy_ids(feature)
    else:
        attr_dict = feature.attributes
    if output_type == 'gff3':
        line = make_gff_line(attr_dict, feature)
    else:
        line = make_gtf_line(attr_dict, feature)
    out_file.write(line)
    out_file.write("\n")


def make_gff_line(attr_dict, feature):
    # Write ID as the first attribute if it exists (for all features including CDS)
    # GFF3 spec allows CDS features to have IDs, and CDS from the same mRNA should have the same ID
    attributes_str = ""
    if "ID" in attr_dict and len(attr_dict["ID"]) > 0:
        attributes_str = "ID=" + attr_dict["ID"][0] + ";"
    
    for attr in attr_dict:
        if attr != "copy_id":
            value_str = ""
            for value in attr_dict[attr]:
                value_str += value + ","
            if attr != "ID":
                attributes_str += (attr + "=" + value_str[:-1] + ";")
    return feature.seqid + "\t" + feature.source + "\t" + feature.featuretype + "\t" + str(feature.start) + \
           "\t" + str(feature.end) + "\t" + "." + "\t" + feature.strand + "\t" + feature.frame + "\t" + attributes_str[:-1]


def edit_copy_ids(feature):
    new_attr_dict = feature.attributes.copy()
    copy_num = feature.attributes["extra_copy_number"][0]

    for attr in feature.attributes:
        try:
            if attr[-3:] == "_id":
                new_attr_dict[attr] = [feature.attributes[attr][0] + "_" + copy_num]
            elif attr == 'ID':
                new_attr_dict["ID"] = [feature.attributes["ID"][0] + "_" + copy_num]
            elif attr == "Parent":
                new_attr_dict["Parent"] = [feature.attributes["Parent"][0] + "_" + copy_num]
        except (KeyError, IndexError) as e:
            print(f"Error processing attribute '{attr}': {e}")
            # Skip any attribute that raises a KeyError or IndexError
            continue
            
    return new_attr_dict

    # for attr in feature.attributes:
    #     if attr[-3:] == "_id":
    #         new_attr_dict[attr] = [feature.attributes[attr][0] + "_" + copy_num]
    #     elif attr == 'ID':
    #         new_attr_dict["ID"] = [feature.attributes["ID"][0]+ "_" + copy_num]
    #     elif attr == "Parent":
    #         new_attr_dict["Parent"] = [feature.attributes["Parent"][0] + "_" + copy_num]
    # return new_attr_dict


def make_gtf_line(attr_dict, feature):
    attributes_str = ""
    for attr in attr_dict:
        if attr != "copy_id":
            if len(attr_dict[attr]) >0:
                value_str = ""
                for value in attr_dict[attr]:
                     value_str += value + ","
                attributes_str += (attr + " " + '"' + value_str[:-1] + '"' + "; ")
    return feature.seqid + "\t" + feature.source + "\t" + feature.featuretype + "\t" + str(feature.start) + \
           "\t" + str(feature.end) + "\t" + "." + "\t" + feature.strand + "\t" + "." + "\t" + attributes_str



import setuptools
from pathlib import Path

this_directory = Path(__file__).resolve().parent
long_description = (this_directory / "./README.md").read_text()
setuptools.setup(
	name="lifton",
	version="1.0.7",
	author="Kuan-Hao Chao",
	author_email="kh.chao@cs.jhu.edu",
	description="Combining DNA and protein alignments to improve genome annotation with LiftOn",
	url="https://github.com/Kuanhao-Chao/Lifton",
	install_requires=['numpy>= 1.22.0', "biopython>=1.76", "cigar>=0.1.3", "parasail>=1.2.4", 'intervaltree>=3.1.0', 'interlap>=0.2.6', 'networkx>=3.3', 'pyfaidx>=0.5.8', 'pysam>=0.19.1', 'gffutils>=0.10.1', 'ujson>=3.2.0', 'pytest>=7.0.0'],
	python_requires='>=3.6',
	packages=setuptools.find_packages(),
	entry_points={'console_scripts': ['lifton = lifton.lifton:main'], },
        long_description=long_description,
        long_description_content_type='text/markdown'
)


import numpy as np


def count_overlap(start1, end1, start2, end2):
    overlap = min(end1, end2) - max(start1, start2) +1
    return overlap


def get_relative_child_coord(parent, coord, is_reverse):
    if is_reverse:
        relative_coord = parent.end - coord
    else:
        relative_coord = coord - parent.start
    return relative_coord


def merge_children_intervals(children):
    if len(children) == 0:
        return []
    intervals = [[child.start, child.end] for child in children]
    intervals.sort(key=lambda interval: interval[0])
    merged = [intervals[0]]
    for current in intervals:
        previous = merged[-1]
        if current[0] <= previous[1]:
            previous[1] = max(previous[1], current[1])
        else:
            merged.append(current)
    return merged


def get_parent_list(feature_list):
    parent_list = []
    for feature in feature_list:
        feature_set = feature_list[feature]
        if len(feature_set) > 0:
            parent_list.append(feature_set[0])
    return parent_list


def clear_scores(feature_list, parent_dict):
    for parent in feature_list:
        for feature in feature_list[parent]:
            if feature.id in parent_dict:
                feature.score = -1
    return feature_list


def find_parent_order(parents):
    parents.sort(key=lambda x: (x.seqid, x.start))
    return np.array([[parent.id, parent] for parent in parents])


def convert_id_to_original(id):
    frag_split = id.split("_frag")[0]
    copy_tag_len = len(frag_split.split("_")[-1])
    original_parent_name = frag_split[:-copy_tag_len - 1]
    return original_parent_name


def get_copy_tag(id):
    copy_tag_len = len(id.split("_")[-1])
    copy_tag = id[-copy_tag_len - 1:]
    return copy_tag


def get_strand(aln, parent):
    if aln.is_reverse:
        if parent.strand == "-":
            strand = '+'
        else:
            strand = "-"
    else:
        strand = parent.strand
    return strand


def find_overlaps(start, end, chrm, strand, feature_name, intervals, parent_dict, lifted_features_list,
                  max_overlap_non_copies):
    all_overlaps = intervals.find((start, end))
    incorrect_overlaps = []
    filtered_overlaps = [overlap for overlap in all_overlaps if
                         overlap[2][1].seqid == chrm and overlap[2][1].strand == strand]

    for overlap in filtered_overlaps:
        shortest_feature_length = min(end - start, overlap[1] - overlap[0]) +1
        overlap_amount = count_overlap(start, end, overlap[0], overlap[1])
        ref_feature = parent_dict[convert_id_to_original(feature_name)]
        ref_overlap_feature = parent_dict[convert_id_to_original(overlap[2][0])]
        if get_copy_tag(feature_name) != '_0' or get_copy_tag(overlap[2][0]) != '_0':
            max_overlap = 0
        else:
            max_overlap = max_overlap_non_copies
        if overlaps_in_ref_annotation(ref_feature, ref_overlap_feature) is False and overlap[2][0] in \
                lifted_features_list and overlap_amount / shortest_feature_length > max_overlap:
            incorrect_overlaps.append(overlap)
    return incorrect_overlaps


def overlaps_in_ref_annotation(ref_feature1, ref_feature2):
    if ref_feature1.seqid != ref_feature2.seqid:
        return False
    if ref_feature1.strand != ref_feature2.strand:
        return False
    if ref_feature1.id == ref_feature2.id:
        return False
    else:
        return count_overlap(ref_feature1.start, ref_feature1.end,
                             ref_feature2.start,
                             ref_feature2.end) > 0


def find_nonoverlapping_upstream_neighbor(parent_order, feature_name):
    feature_location = np.where(parent_order[:, 0] == feature_name)[0][0]
    neighbor_indx = feature_location - 1
    neighbor = parent_order[neighbor_indx][1]
    neighbor_id = parent_order[neighbor_indx][0]
    feature = parent_order[neighbor_indx + 1][1]
    if neighbor.seqid != feature.seqid or neighbor_indx < 0:
        return None
    else:
        return neighbor_id



from lifton.liftoff  import fix_overlapping_features, lift_features, liftoff_utils, align_features, extract_features


def lift_original_annotation(ref_chroms, target_chroms, lifted_features_list, args, unmapped_features, parents_to_lift, ref_db):
    liftover_type = "chrm_by_chrm"
    if target_chroms[0] == args.target and args.exclude_partial == False:
        min_cov, min_seqid = 0.05, 0.05
    else:
        min_cov, min_seqid = args.a, args.s

    feature_hierarchy, feature_db, ref_parent_order = extract_features.extract_features_to_lift(ref_chroms,
                                                                                                liftover_type,
                                                                                                parents_to_lift, ref_db, args)

    align_and_lift_features(ref_chroms, target_chroms, args, feature_hierarchy, liftover_type, unmapped_features,
                            feature_db, lifted_features_list, ref_parent_order, min_cov, min_seqid, args.overlap)

    return feature_db, feature_hierarchy, ref_parent_order


def align_and_lift_features(ref_chroms, target_chroms, args, feature_hierarchy, liftover_type, unmapped_features,
                            feature_db, lifted_features_list, ref_parent_order, min_cov, min_seqid, max_overlap):
    aligned_segments= align_features.align_features_to_target(ref_chroms, target_chroms, args,
                                                                      feature_hierarchy,
                                                               liftover_type, unmapped_features)
    print("lifting features")
    feature_locations = None
    lift_features.lift_all_features(aligned_segments, min_cov, feature_db,  feature_hierarchy,
                                    unmapped_features, lifted_features_list, min_seqid, feature_locations, args,
                                    ref_parent_order)

    fix_overlapping_features.fix_incorrectly_overlapping_features(lifted_features_list, lifted_features_list,
                                                                  aligned_segments, unmapped_features,
                                                                  min_cov, feature_hierarchy,
                                                                  feature_db, ref_parent_order, min_seqid, args,
                                                                   max_overlap)


def map_unmapped_genes_agaisnt_all(unmapped_features, ref_chroms, target_chroms, lifted_features_list, feature_db,
                                   feature_hierarchy, ref_parent_order, args):
    liftoff_utils.clear_scores(lifted_features_list, feature_hierarchy.parents)
    unmapped_dict = get_unmapped_genes(unmapped_features)
    if args.exclude_partial:
        min_cov, min_seqid = args.a, args.s
    else:
        min_cov, min_seqid = 0.05, 0.05
    liftover_type = "unmapped"
    extract_features.get_gene_sequences(unmapped_dict, ref_chroms, args, liftover_type)
    unmapped_features = []
    align_and_lift_features(ref_chroms, target_chroms, args, feature_hierarchy, liftover_type, unmapped_features,
                            feature_db, lifted_features_list, ref_parent_order, min_cov, min_seqid, args.overlap,)
    return unmapped_features


def get_unmapped_genes(unmapped_features):
    unmapped_dict = {}
    for feature in unmapped_features:
        unmapped_dict[feature.id] = feature
    return unmapped_dict


def map_unplaced_genes(unmapped_features, ref_chroms, target_chroms,
                       lifted_features_list, feature_db, feature_hierarchy, ref_parent_order, args):
    liftoff_utils.clear_scores(lifted_features_list, feature_hierarchy.parents)
    liftover_type = "unplaced"
    unplaced_dict = get_features_from_unplaced_seq(ref_chroms, feature_hierarchy)
    extract_features.get_gene_sequences(unplaced_dict, ref_chroms, args, liftover_type)
    if args.exclude_partial:
        min_cov, min_seqid = args.a, args.s
    else:
        min_cov, min_seqid = 0.05, 0.05
    align_and_lift_features(ref_chroms, target_chroms, args, feature_hierarchy, liftover_type, unmapped_features,
                            feature_db, lifted_features_list, ref_parent_order, min_cov, min_seqid,args.overlap)


def get_features_from_unplaced_seq(ref_chroms, feature_hierarchy):
    unplaced_dict = {}
    for feature_name in feature_hierarchy.parents:
        feature = feature_hierarchy.parents[feature_name]
        if feature.seqid in ref_chroms:
            unplaced_dict[feature.id] = feature
    return unplaced_dict


def map_extra_copies(ref_chroms, target_chroms, lifted_features_list, feature_hierarchy, feature_db, ref_parent_order,
                     args):
    liftoff_utils.clear_scores(lifted_features_list, feature_hierarchy.parents)
    unmapped_features = []
    liftover_type = "copies"
    extract_features.get_gene_sequences(feature_hierarchy.parents, ref_chroms, args, liftover_type)
    min_cov, min_seqid = 0, args.sc
    align_and_lift_features(ref_chroms, target_chroms, args, feature_hierarchy, liftover_type, unmapped_features,
                            feature_db, lifted_features_list, ref_parent_order, min_cov, min_seqid, args.overlap)



class new_feature:
    def __init__(self, id, featuretype, seqid, source, strand, start, end, frame, attributes):
        self.id = id
        self.featuretype = featuretype
        self.seqid = seqid
        self.source = source
        self.strand = strand
        self.start = start
        self.end = end
        self.frame = frame
        self.attributes= attributes



class aligned_seg:
    def __init__(self, aln_id, query_name, reference_name, query_block_start, query_block_end, reference_block_start,
                 reference_block_end, is_reverse, mismatches):
        self.aln_id = aln_id
        self.query_name = query_name
        self.reference_name = reference_name
        self.query_block_start = query_block_start
        self.query_block_end = query_block_end
        self.is_reverse = is_reverse
        self.reference_block_start = reference_block_start
        self.reference_block_end = reference_block_end
        self.mismatches = mismatches



import sys
from lifton import logger

def print_report(ref_features_dict, transcripts_stats_dict, fw_unmapped, fw_extra_copy, fw_mapped_feature, fw_mapped_trans, debug=False):
    """
    This function prints the report of the lifton results.

    Parameters:
    - ref_features_dict: reference features dictionary
    - fw_unmapped: file writer for unmapped features
    - fw_extra_copy: file writer for extra copy features
    - fw_mapped_feature: file writer for mapped features
    - fw_mapped_trans: file writer for mapped transcripts
    - debug: debug mode
    """
    LIFTED_FEATURES = 0
    MISSED_FEATURES = 0
    LIFTED_SINGLE_CODING_FEATURES = 0
    LIFTED_SINGLE_NONCODING_FEATURES = 0
    LIFTED_SINGLE_OTHER_FEATURES = 0
    LIFTED_EXTRA_CODING_FEATURES = 0
    LIFTED_EXTRA_NONCODING_FEATURES = 0
    LIFTED_EXTRA_OTHER_FEATURES = 0
    LIFTED_EXTRA_CODING_SUM_FEATURES = 0
    LIFTED_EXTRA_NONCODING_SUM_FEATURES = 0
    LIFTED_EXTRA_OTHER_SUM_FEATURES = 0
    
    # Feature stats (gene)
    for feature in ref_features_dict.keys():
        if feature == "LiftOn-gene":
            continue
        copy_num = ref_features_dict[feature].copy_num
        TYPE = ""
        if copy_num >= 1:
            LIFTED_FEATURES += 1
            if ref_features_dict[feature].is_protein_coding:
                TYPE = "coding"
                if copy_num == 1:
                    LIFTED_SINGLE_CODING_FEATURES += 1
                else:
                    LIFTED_EXTRA_CODING_FEATURES += 1
                    LIFTED_EXTRA_CODING_SUM_FEATURES += copy_num
            elif ref_features_dict[feature].is_non_coding:
                TYPE = "non-coding"
                if copy_num == 1:
                    LIFTED_SINGLE_NONCODING_FEATURES += 1
                else:
                    LIFTED_EXTRA_NONCODING_FEATURES += 1
                    LIFTED_EXTRA_NONCODING_SUM_FEATURES += copy_num
            else:
                TYPE = "other"
                if copy_num == 1:
                    LIFTED_SINGLE_OTHER_FEATURES += 1
                else:
                    LIFTED_EXTRA_OTHER_FEATURES += 1
                    LIFTED_EXTRA_OTHER_SUM_FEATURES += copy_num

            fw_mapped_feature.write(f"{feature}\t{copy_num}\t{TYPE}\n")
            if copy_num > 1:
                fw_extra_copy.write(f"{feature}\t{copy_num}\t{TYPE}\n")
        elif copy_num == 0:
            MISSED_FEATURES += 1
            if ref_features_dict[feature].is_protein_coding:
                TYPE = "coding"
            elif ref_features_dict[feature].is_non_coding:
                TYPE = "non-coding"
            else:
                TYPE = "other"
            fw_unmapped.write(f"{feature}\t{TYPE}\n")

    # Transcript stats        
    for TYPE, transs in transcripts_stats_dict.items():
        for trans, trans_copy_num in transs.items():
            fw_mapped_trans.write(f"{trans}\t{trans_copy_num}\t{TYPE}\n")

    print("\n\n*********************************************", file=sys.stderr)
    print(f"* Total features in reference\t\t: {len(ref_features_dict.keys())-1}", file=sys.stderr)
    print(f"* Lifted feature\t\t\t: {LIFTED_FEATURES} ({LIFTED_SINGLE_CODING_FEATURES + LIFTED_EXTRA_CODING_FEATURES} + {LIFTED_SINGLE_NONCODING_FEATURES + LIFTED_EXTRA_NONCODING_FEATURES} + {LIFTED_SINGLE_OTHER_FEATURES + LIFTED_EXTRA_OTHER_FEATURES})", file=sys.stderr)
    print(f"\t* Protein-coding feature\t: {LIFTED_SINGLE_CODING_FEATURES + LIFTED_EXTRA_CODING_FEATURES}", file=sys.stderr)
    print(f"\t* Non-coding feature\t\t: {LIFTED_SINGLE_NONCODING_FEATURES + LIFTED_EXTRA_NONCODING_FEATURES}", file=sys.stderr)
    print(f"\t* Other feature\t\t\t: {LIFTED_SINGLE_OTHER_FEATURES + LIFTED_EXTRA_OTHER_FEATURES}", file=sys.stderr)
    print(f"* Missed feature\t\t\t: {MISSED_FEATURES}\n", file=sys.stderr)

    print(f"* Total features in target\t\t: {LIFTED_SINGLE_CODING_FEATURES + LIFTED_EXTRA_CODING_SUM_FEATURES + LIFTED_SINGLE_NONCODING_FEATURES + LIFTED_EXTRA_NONCODING_SUM_FEATURES + LIFTED_SINGLE_OTHER_FEATURES + LIFTED_EXTRA_OTHER_SUM_FEATURES} ({LIFTED_SINGLE_CODING_FEATURES + LIFTED_EXTRA_CODING_SUM_FEATURES} + {LIFTED_SINGLE_NONCODING_FEATURES + LIFTED_EXTRA_NONCODING_SUM_FEATURES} + {LIFTED_SINGLE_OTHER_FEATURES + LIFTED_EXTRA_OTHER_SUM_FEATURES})", file=sys.stderr)
    print(f"\t* Protein-coding feature\t: {LIFTED_SINGLE_CODING_FEATURES + LIFTED_EXTRA_CODING_SUM_FEATURES} ({LIFTED_SINGLE_CODING_FEATURES} + {LIFTED_EXTRA_CODING_SUM_FEATURES})", file=sys.stderr)
    print(f"\t\t* single copy\t\t: {LIFTED_SINGLE_CODING_FEATURES}", file=sys.stderr)
    print(f"\t\t* > 1 copy\t\t: {LIFTED_EXTRA_CODING_FEATURES}, {LIFTED_EXTRA_CODING_SUM_FEATURES} in total", file=sys.stderr)
    print(f"\t* Non-coding feature\t\t: {LIFTED_SINGLE_NONCODING_FEATURES + LIFTED_EXTRA_NONCODING_SUM_FEATURES} ({LIFTED_SINGLE_NONCODING_FEATURES} + {LIFTED_EXTRA_NONCODING_SUM_FEATURES})", file=sys.stderr)
    print(f"\t\t* single copy\t\t: {LIFTED_SINGLE_NONCODING_FEATURES}", file=sys.stderr)
    print(f"\t\t* > 1 copy\t\t: {LIFTED_EXTRA_NONCODING_FEATURES}, {LIFTED_EXTRA_NONCODING_SUM_FEATURES} in total", file=sys.stderr)
    print(f"\t* Other feature\t\t\t: {LIFTED_SINGLE_OTHER_FEATURES + LIFTED_EXTRA_OTHER_SUM_FEATURES} ({LIFTED_SINGLE_OTHER_FEATURES} + {LIFTED_EXTRA_OTHER_SUM_FEATURES})", file=sys.stderr)
    print(f"\t\t* single copy\t\t: {LIFTED_SINGLE_OTHER_FEATURES}", file=sys.stderr)
    print(f"\t\t* > 1 copy\t\t: {LIFTED_EXTRA_OTHER_FEATURES}, {LIFTED_EXTRA_OTHER_SUM_FEATURES} in total", file=sys.stderr)
    print(f"*********************************************")



import gffutils
import sys
import os
import subprocess
from collections import defaultdict
import re
from lifton import extract_sequence


class Annotation():

    def __init__(self, file_name, infer_genes, infer_transcripts, merge_strategy, id_spec, force, verbose, auto_convert_gtf=True):
        self.file_name = file_name
        self.merge_strategy = merge_strategy
        self.id_spec = id_spec
        self.force = force
        self.verbose = verbose
        self.auto_convert_gtf = auto_convert_gtf
        
        # Detect file format and adjust inference settings
        # Default is GFF3 format (most common and recommended)
        file_format = self.detect_file_format()
        
        if file_format == "GTF format":
            print(f"INFO: Detected GTF format for {file_name}")
            print("INFO: GTF format support is experimental. For best results, consider converting to GFF3 using:")
            print("  agat_sp_gtf2gff.pl --gtf input.gtf -o output.gff3")
            print("  or")
            print("  gffread -E input.gtf -o output.gff3")
            # For GTF files, we need to infer genes and transcripts
            # GTF files don't have explicit gene/transcript features, only exons/CDS with gene_id/transcript_id
            # So we always enable inference for GTF files (user can't really override this)
            self.infer_genes = True
            self.infer_transcripts = True
            if self.verbose:
                print("  Automatically enabling gene and transcript inference for GTF format")
            
            # Optionally convert GTF to GFF3 for better compatibility
            if self.auto_convert_gtf:
                converted_file = self.convert_gtf_to_gff3()
                if converted_file:
                    # Verify the converted file exists and is valid
                    if os.path.exists(converted_file) and os.path.getsize(converted_file) > 0:
                        self.file_name = converted_file
                        # Re-detect format to ensure it's now GFF3
                        converted_format = self.detect_file_format()
                        if converted_format == "GFF format":
                            if self.verbose:
                                print(f"  Successfully converted GTF to GFF3: {converted_file}")
                        else:
                            if self.verbose:
                                print(f"  Warning: Converted file may not be valid GFF3 format")
                    else:
                        if self.verbose:
                            print(f"  Warning: Conversion produced empty or invalid file, using original GTF")
        elif file_format == "Unknown format":
            # If format detection fails, default to GFF3 and use user-provided settings
            if self.verbose:
                print(f"Warning: Could not determine file format for {file_name}, assuming GFF3 format")
            self.infer_genes = infer_genes
            self.infer_transcripts = infer_transcripts
        else:
            # For GFF3 files (or any other recognized format), use user-provided settings
            # argparse passes False when flags are not provided, True when they are
            self.infer_genes = infer_genes
            self.infer_transcripts = infer_transcripts
        
        self.get_db_connnection()
        
    def get_db_connnection(self):
        try:
            feature_db = gffutils.FeatureDB(self.file_name)
        except:
            feature_db = self.build_database()
        self.db_connection = feature_db


    def build_database(self):
        if self.infer_genes:
            disable_genes = False
        else:
            disable_genes = True
        if self.infer_transcripts:
            disable_transcripts = False
        else:
            disable_transcripts = True
        try:
            transform_func = self.get_transform_func()
            feature_db = gffutils.create_db(self.file_name, self.file_name + "_db", 
                                        merge_strategy=self.merge_strategy, 
                                        id_spec=self.id_spec,
                                        force=self.force,
                                        verbose=self.verbose, disable_infer_transcripts=disable_transcripts,
                                            disable_infer_genes=disable_genes, transform=transform_func)
        except Exception as e:
            error_msg = str(e)
            if "UNIQUE constraint failed" in error_msg or "duplicate" in error_msg.lower():
                print(f"Warning: Duplicate feature IDs detected in {self.file_name}. Attempting to handle duplicates...")
            else:
                print(f"gffutils database build failed with: {e}")
            feature_db = self.build_database_again()
            return feature_db


    def build_database_again(self):
        if self.infer_genes:
            disable_genes = False
        else:
            disable_genes = True
        if self.infer_transcripts:
            disable_transcripts = False
        else:
            disable_transcripts = True
        # For duplicate IDs (especially non-overlapping CDS features), use unique ID transformation
        # This is better than 'merge' strategy which is designed for overlapping features
        try:
            transform_func = self.get_unique_id_transform()
            feature_db = gffutils.create_db(self.file_name, self.file_name + "_db", 
                                        merge_strategy="create_unique",
                                        id_spec=self.id_spec,
                                        force=True,
                                        verbose=self.verbose, disable_infer_transcripts=disable_transcripts,
                                            disable_infer_genes=disable_genes, transform=transform_func)
            print("Successfully built database using unique ID transformation to handle duplicate IDs.")
            return feature_db
        except Exception as e1:
            print(f"Unique ID transformation failed: {e1}")
            print("Attempting to build database with merge strategy...")
            # If transform fails, try merge strategy as a fallback
            try:
                transform_func = self.get_transform_func()
                feature_db = gffutils.create_db(self.file_name, self.file_name + "_db", 
                                            merge_strategy="merge",
                                            id_spec=self.id_spec,
                                            force=True,
                                            verbose=self.verbose, disable_infer_transcripts=disable_transcripts,
                                                disable_infer_genes=disable_genes, transform=transform_func)
                print("Successfully built database using merge strategy to handle duplicate IDs.")
                return feature_db
            except Exception as e2:
                print(f"ERROR: Failed to build database after trying multiple strategies.")
                print(f"Last error: {e2}")
                print(f"Please check the GFF file {self.file_name} for duplicate feature IDs or other issues.")
                sys.exit(1)


    def get_transform_func(self):
        if self.infer_genes is False:
            return None
        else:
            return transform_func

    def get_unique_id_transform(self):
        """Create a transform function that ensures unique IDs by appending a counter to duplicates.
        This is especially important for non-overlapping CDS features in RefSeq annotations that may have duplicate IDs.
        """
        # Dictionary to track how many times we've seen each ID
        seen_ids = {}
        
        def unique_id_transform(feature):
            # Apply the original transform if needed
            if self.infer_genes:
                feature = transform_func(feature)
            
            # Get the ID from the feature - gffutils uses the 'id' property or ID attribute
            # We need to check both the id property and the ID attribute
            original_id = None
            if hasattr(feature, 'id') and feature.id:
                original_id = feature.id
            elif 'ID' in feature.attributes and len(feature.attributes['ID']) > 0:
                original_id = feature.attributes['ID'][0]
            
            # If no ID, gffutils will generate one, so we don't need to modify
            if not original_id:
                return feature
            
            # Track seen IDs and make unique if needed
            # First occurrence keeps original ID, subsequent ones get _dup1, _dup2, etc.
            if original_id in seen_ids:
                seen_ids[original_id] += 1
                new_id = f"{original_id}_dup{seen_ids[original_id]}"
                # Update both the ID attribute and the id property
                # This ensures gffutils uses the new unique ID
                if 'ID' in feature.attributes:
                    feature.attributes['ID'] = [new_id]
                if hasattr(feature, 'id'):
                    feature.id = new_id
            else:
                # First time seeing this ID - mark it as seen but don't modify
                seen_ids[original_id] = 0
            
            return feature
        
        return unique_id_transform



    def get_protein_coding_features(self, feature_types):
        protein_coding_genes = []
        for f_type in feature_types:
            for feature in self.db_connection.features_of_type(featuretype=f_type):
                if self.is_highest_parent(feature):
                    CDS_list = [child for child in self.db_connection.children(feature) if child.featuretype == 'CDS']
                    if len(CDS_list) > 0:
                        protein_coding_genes.append(feature.id)
        return protein_coding_genes


    def get_noncoding_features(self, feature_types):
        noncoding_genes = []
        for f_type in feature_types:
            for feature in self.db_connection.features_of_type(featuretype=f_type):
                if self.is_highest_parent(feature):
                    CDS_list = [child for child in self.db_connection.children(feature) if child.featuretype == 'CDS']
                    if len(CDS_list)==  0:
                        noncoding_genes.append(feature.id)
        return noncoding_genes


    def get_novel_protein_coding_features(self, ref_genes, feature_types):
        novel_protein_coding = []
        all_protein_coding = self.get_protein_coding_features(feature_types)
        for feature in all_protein_coding:
            if feature not in ref_genes:
                novel_protein_coding.append(feature)
        return novel_protein_coding


    def get_novel_noncoding_features(self,ref_genes, feature_types):
        novel_noncoding = []
        all_noncoding = self.get_noncoding_features(feature_types)
        for feature in all_noncoding:
            if feature not in ref_genes:
                novel_noncoding.append(feature)
        return novel_noncoding


    def get_all_parent_feature_ids(self, feature_types):
        feature_ids = []
        for f_type in feature_types:
            feature_ids += [feature.id for feature in self.db_connection.features_of_type(
                featuretype=f_type) if self.is_highest_parent(feature)]
        return feature_ids


    def make_parent_to_child_dict(self, protein_coding, gene):
        child_dict=defaultdict(list)
        child_list = self.db_connection.children(gene)
        child_count = 0
        for child in child_list:
            child_count += 1
            if (protein_coding and child.featuretype == "CDS") or (protein_coding is False and self.is_lowest_child(
                    child)):
                parent = list(self.db_connection.parents(child, level=1))[0]
                child_dict[parent.id].append(child)
        if child_count == 0:
            child_dict[gene] = [self.db_connection[gene]]
        return child_dict


    def get_features_of_type(self,feature_types):
        features_of_type = []
        for f_type in feature_types:
            features_of_type += list(self.db_connection.features_of_type(f_type))
        return features_of_type


    def get_feature_dict(self, feature_types):
        id_to_feature = {}
        features = self.get_features_of_type(feature_types)
        for feature in features:
            id_to_feature[feature.id] = feature
        return id_to_feature


    def get_source_name(self, feature_name):
        feature = self.db_connection[feature_name]
        if 'extra_copy_number' in feature.attributes:
            if feature.attributes['extra_copy_number'][0] == feature.id.split("_")[-1]:
                return "_".join(feature.id.split("_")[:-1])
        return feature.id


    def is_lowest_child(self, feature_name):
        return len(list(self.db_connection.children(feature_name))) == 0


    def is_highest_parent(self, feature_name):
        return len(list(self.db_connection.parents(feature_name))) == 0


    def get_paralog_name(self, feature_name):
        gene_attributes = self.db_connection[feature_name].attributes
        if "extra_copy_number" in gene_attributes:
            paralog_name = "_".join(gene_attributes["ID"][0].split("_")[:-1])
            return paralog_name
        return ""


    def get_num_levels(self, feature_name):
        level = 1
        new_children = list(self.db_connection.children(feature_name, level=level))
        while len(new_children) !=0:
            level += 1
            new_children = list(self.db_connection.children(feature_name, level=level))
        return level

    def detect_file_format(self):
        """Detect if the input file is GTF or GFF3 format."""
        try:
            return extract_sequence.determine_file_format(self.file_name)
        except Exception as e:
            if self.verbose:
                print(f"Warning: Could not detect file format: {e}")
            # Default to GFF3 if detection fails
            return "GFF format"

    def convert_gtf_to_gff3(self):
        """Convert GTF file to GFF3 format using agat or gffread.
        Returns the path to the converted file, or None if conversion fails.
        """
        # Check if agat is available (try different agat commands)
        agat_available = False
        agat_cmd = None
        for cmd_name in ["agat_convert_sp_gff2gtf.pl", "agat_sp_gtf2gff.pl", "agat"]:
            if self._check_tool_available(cmd_name):
                agat_available = True
                agat_cmd = cmd_name
                break
        
        gffread_available = self._check_tool_available("gffread")
        
        if not agat_available and not gffread_available:
            if self.verbose:
                print("  Warning: Neither agat nor gffread found. GTF file will be used directly.")
                print("  For best results with GTF files, consider converting to GFF3 using:")
                print("    agat_sp_gtf2gff.pl --gtf input.gtf -o output.gff3")
                print("  or")
                print("    gffread -E input.gtf -o output.gff3")
            return None
        
        # Create output filename
        base_name = os.path.splitext(self.file_name)[0]
        output_file = base_name + "_converted.gff3"
        
        # Try gffread first (more commonly available and reliable)
        if gffread_available:
            try:
                cmd = ["gffread", "-E", self.file_name, "-o", output_file]
                result = subprocess.run(cmd, capture_output=True, text=True, check=True)
                if os.path.exists(output_file) and os.path.getsize(output_file) > 0:
                    return output_file
            except (subprocess.CalledProcessError, FileNotFoundError) as e:
                if self.verbose:
                    print(f"  Warning: gffread conversion failed: {e}")
        
        # Try agat as fallback
        if agat_available:
            try:
                # Try different agat command formats
                if "gtf2gff" in agat_cmd:
                    cmd = [agat_cmd, "--gtf", self.file_name, "-o", output_file]
                else:
                    # Generic agat command - may need adjustment
                    cmd = [agat_cmd, "--gff", self.file_name, "-o", output_file]
                result = subprocess.run(cmd, capture_output=True, text=True, check=True)
                if os.path.exists(output_file) and os.path.getsize(output_file) > 0:
                    return output_file
            except (subprocess.CalledProcessError, FileNotFoundError) as e:
                if self.verbose:
                    print(f"  Warning: agat conversion failed: {e}")
        
        return None

    def _check_tool_available(self, tool_name):
        """Check if a command-line tool is available."""
        try:
            subprocess.run(["which", tool_name], capture_output=True, check=True)
            return True
        except (subprocess.CalledProcessError, FileNotFoundError):
            return False



def transform_func(x):
    if 'transcript_id' in x.attributes:
        x.attributes['transcript_id'][0] += '_transcript'
    return x


def get_perc_id(feature):
    if 'sequence_ID' in feature.attributes:
        return float(feature.attributes['sequence_ID'][0])
    return 0.0


def merge_children_intervals(children):
    if len(children) == 0:
        return []
    intervals = [[child.start, child.end] for child in children]
    intervals.sort(key=lambda interval: interval[0])
    merged = [intervals[0]]
    for current in intervals:
        previous = merged[-1]
        if current[0] <= previous[1]:
            previous[1] = max(previous[1], current[1])
        else:
            merged.append(current)
    return merged